from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.enums import (
    RatingEstimatorFeatures,
    RatingHistoricalFeatures,
)

from player_performance_ratings.data_structures import (
    Match,
    PlayerRating,
    TeamRating,
    ColumnNames,
)


class RatingGenerator(ABC):

    def __init__(
        self,
        performance_column: str,
        estimator_features_pass_through: Optional[list[RatingEstimatorFeatures]],
        historical_features_out: Optional[list[RatingHistoricalFeatures]],
        seperate_player_by_position: Optional[bool] = False,
    ):
        self.performance_column = performance_column
        self.seperate_player_by_position = seperate_player_by_position
        self._estimator_features_out = []
        self._historical_features_out = historical_features_out or []
        self._estimator_features_pass_through = estimator_features_pass_through
        self._ratings_df = None
        self.column_names = None

    def reset_ratings(self):
        self._ratings_df = None

    @abstractmethod
    def generate_historical(
        self,
        df: Optional[pd.DataFrame],
        column_names: ColumnNames,
        matches: Optional[list[Match]] = None,
    ) -> dict[RatingEstimatorFeatures, list[float]]:
        pass

    @abstractmethod
    def generate_future(
        self, df: Optional[pd.DataFrame], matches: Optional[list[Match]] = None
    ) -> dict[RatingEstimatorFeatures, list[float]]:
        pass

    @property
    @abstractmethod
    def player_ratings(self) -> dict[str, PlayerRating]:
        pass

    @property
    @abstractmethod
    def team_ratings(self) -> list[TeamRating]:
        pass

    @property
    def estimator_features_out(self) -> list[RatingEstimatorFeatures]:
        return self._estimator_features_out

    @property
    def features_out(
        self,
    ) -> list[Union[RatingEstimatorFeatures, RatingHistoricalFeatures]]:
        return self._estimator_features_out + self._historical_features_out

    @property
    def estimator_features_return(self) -> list[RatingEstimatorFeatures]:
        if self._estimator_features_pass_through:
            return list(
                set(self._estimator_features_pass_through + self.estimator_features_out)
            )
        return self.estimator_features_out

    @property
    def ratings_df(self) -> pd.DataFrame:
        return self._ratings_df



import logging
import math
from typing import Dict, Optional, Tuple

from player_performance_ratings.data_structures import (
    MatchPlayer,
    PlayerRating,
    PreMatchTeamRating,
    PreMatchPlayerRating,
    PlayerRatingChange,
    Team,
    MatchTeam,
    TeamRatingChange,
)
from player_performance_ratings.ratings.rating_calculators.performance_predictor import (
    RatingDifferencePerformancePredictor,
    PerformancePredictor,
)
from player_performance_ratings.ratings.rating_calculators.start_rating_generator import (
    StartRatingGenerator,
)

MATCH_CONTRIBUTION_TO_SUM_VALUE = 1
EXPECTED_MEAN_CONFIDENCE_SUM = 30


class MatchRatingGenerator:

    def __init__(
        self,
        start_rating_generator: Optional[StartRatingGenerator] = None,
        performance_predictor: Optional[PerformancePredictor] = None,
        rating_change_multiplier: float = 50,
        confidence_days_ago_multiplier: float = 0.06,
        confidence_max_days: int = 90,
        confidence_value_denom: float = 140,
        confidence_max_sum: float = 150,
        confidence_weight: float = 0.9,
        min_rating_change_multiplier_ratio: float = 0.1,
        league_rating_change_update_threshold: float = 100,
        league_rating_adjustor_multiplier: float = 0.05,
        team_id_change_confidence_sum_decrease: float = 3,
    ):
        """
        :param start_rating_generator:
            Class object for generating start rating (the rating a player receives in its very first match).

        :param performance_predictor:
            Class object for predicting performance of a player in a match.
        This is used to adjust for the current quality of opponents faced or other contexts dependant on the class used.

        :param rating_change_multiplier:
            The base multiplier for how much a player's rating changes after a match.
            Equivalent of the K-factor in Elo.


        :param confidence_days_ago_multiplier:
            Determinutes how much the confidence_sum is affected by how long ago the last match was.
            If a player has not played in a long time, the confidence in a players rating declines.
            This means his future rating changes faster. A higher confidence_days_ago_multiplier  therefore leads to more volatile future rating changes.

        :param confidence_max_days:
            A players confidence_sum decreases by confidence_days_ago_multiplier * min(confidence_max_days, days_ago_since_last_match).
            confidence_max_days therefore puts a limit on how much a players confidence_sum can decrease.
            This ensures a non-linear relationship between inactivity and our confidence in a players rating.

        :param confidence_value_denom:
            Higher confidence_value_denom results in a more volatile rating change multipler based on the confidence_sum.

        :param confidence_max_sum:
            The maximum confidence_sum a player can have.
             If confidence_max_sum is very high, a player with many historical games, can be inactive for a while and still have a high confidence_sum.
             Thus a lower confidence_max_sum increases the speed at which a players confidence_sum decreases when inactive.


        :param min_rating_change_multiplier_ratio:
            To ensure that a player's rating changes at least a little bit after a match,
            min_rating_change_multiplier_ratio ensures that the applied_rating_change_multiplier is at least a certain ratio of the rating_change_multiplier.

        :param confidence_weight:
            Determines how much the applied rating change multiplier is affected by the confidence_sum.
            The confidence_sum being how much data (and how recent) we have on a player.
            If the confidence_ratio is 0, the applied rating_change_multipler is equal to the rating_change_multiplier.
            The higher ratio it is, the more volatile the applied rating_change_multiplier is based on the data we have on a player.
            Recommended is between 0.7 and 1.0

        :param league_rating_change_update_threshold:
            All players within a league gets their rating updated when the sum of all rating changes within a league exceeds this threshold.
            A lower value means that the ratings within a league are updated more often but is more computationally expensive.

        :param league_rating_adjustor_multiplier:
            The amount of which the players within a league gets their rating updated with, is league_rating_adjustor_multiplier * mean_rating_change.
            A higher value makes the performance of other players within a league have a larger impact on a players rating - even if the player hasn't played a match.

        """

        self.confidence_weight = confidence_weight
        self.confidence_days_ago_multiplier = confidence_days_ago_multiplier
        self.confidence_value_denom = confidence_value_denom
        self.min_rating_change_multiplier_ratio = min_rating_change_multiplier_ratio
        self.rating_change_multiplier = rating_change_multiplier
        self.confidence_max_sum = confidence_max_sum
        self.league_rating_adjustor_multiplier = league_rating_adjustor_multiplier
        self.league_rating_change_update_threshold = (
            league_rating_change_update_threshold
        )
        self.confidence_max_days = confidence_max_days
        self.player_ratings: Dict[str, PlayerRating] = {}
        self._league_rating_changes: dict[Optional[str], float] = {}
        self._league_rating_changes_count: dict[str, float] = {}
        self.performance_predictor = (
            performance_predictor or RatingDifferencePerformancePredictor()
        )
        self.start_rating_generator = start_rating_generator or StartRatingGenerator()
        self.team_id_change_confidence_sum_decrease = (
            team_id_change_confidence_sum_decrease
        )

        self._teams: dict[str, Team] = {}

    def generate_pre_match_team_rating(
        self, day_number: int, match_team: MatchTeam
    ) -> PreMatchTeamRating:

        self._teams[match_team.id] = Team(
            id=match_team.id,
            last_match_day_number=day_number,
            player_ids=[p.id for p in match_team.players],
        )

        (
            existing_pre_match_player_ratings,
            pre_match_player_rating_values,
            new_players,
        ) = self._get_pre_match_player_ratings_and_new_players(team=match_team)

        new_player_pre_match_ratings = self._generate_new_player_pre_match_ratings(
            day_number=day_number,
            new_players=new_players,
            team_pre_match_player_ratings=existing_pre_match_player_ratings,
        )
        if (
            len(new_player_pre_match_ratings) > 0
            and len(existing_pre_match_player_ratings) > 0
        ):

            pre_match_player_ratings = []
            for match_player in match_team.players:

                pre_match_player = [
                    p for p in new_player_pre_match_ratings if p.id == match_player.id
                ]
                if len(pre_match_player) == 0:
                    pre_match_player = [
                        p
                        for p in existing_pre_match_player_ratings
                        if p.id == match_player.id
                    ]

                pre_match_player_ratings.append(pre_match_player[0])

        elif len(new_player_pre_match_ratings) == 0:
            pre_match_player_ratings = existing_pre_match_player_ratings
        else:
            pre_match_player_ratings = new_player_pre_match_ratings

        # pre_match_player_ratings += new_player_pre_match_ratings
        pre_match_team_rating_value = self._generate_pre_match_team_rating_value(
            pre_match_player_ratings=pre_match_player_ratings
        )
        pre_match_team_rating_projected_value = (
            self._generate_pre_match_team_rating_projected_value(
                pre_match_player_ratings=pre_match_player_ratings
            )
        )

        return PreMatchTeamRating(
            id=match_team.id,
            players=pre_match_player_ratings,
            rating_value=pre_match_team_rating_value,
            projected_rating_value=pre_match_team_rating_projected_value,
            league=match_team.league,
        )

    def _get_pre_match_player_ratings_and_new_players(
        self, team: MatchTeam
    ) -> Tuple[list[PreMatchPlayerRating], list[float], list[MatchPlayer]]:

        pre_match_player_ratings = []
        player_count = 0

        new_match_entities = []
        pre_match_player_ratings_values = []

        for match_player in team.players:
            if match_player.id in self.player_ratings:

                player_rating = self._get_rating_by_id(id=match_player.id)

                pre_match_player_rating = PreMatchPlayerRating(
                    id=match_player.id,
                    rating_value=player_rating.rating_value,
                    match_performance=match_player.performance,
                    games_played=player_rating.games_played,
                    league=match_player.league,
                    position=match_player.position,
                    other=match_player.others,
                )

            else:
                new_match_entities.append(match_player)
                continue

            player_count += self._get_rating_by_id(match_player.id).games_played

            pre_match_player_ratings.append(pre_match_player_rating)
            pre_match_player_ratings_values.append(pre_match_player_rating.rating_value)

        return (
            pre_match_player_ratings,
            pre_match_player_ratings_values,
            new_match_entities,
        )

    def _generate_new_player_pre_match_ratings(
        self,
        day_number: int,
        new_players: list[MatchPlayer],
        team_pre_match_player_ratings: list[PreMatchPlayerRating],
    ) -> list[PreMatchPlayerRating]:

        pre_match_player_ratings = []

        for match_player in new_players:
            id = match_player.id

            rating_value = self.start_rating_generator.generate_rating_value(
                day_number=day_number,
                match_player=match_player,
                team_pre_match_player_ratings=team_pre_match_player_ratings,
            )

            self.player_ratings[match_player.id] = PlayerRating(
                id=id, rating_value=rating_value
            )

            pre_match_player_rating = PreMatchPlayerRating(
                id=id,
                rating_value=rating_value,
                match_performance=match_player.performance,
                games_played=self.player_ratings[match_player.id].games_played,
                league=match_player.league,
                position=match_player.position,
                other=match_player.others,
            )

            pre_match_player_ratings.append(pre_match_player_rating)

        return pre_match_player_ratings

    def _generate_pre_match_team_rating_projected_value(
        self, pre_match_player_ratings: list[PreMatchPlayerRating]
    ) -> float:
        team_rating = sum(
            player.rating_value
            * player.match_performance.projected_participation_weight
            for player in pre_match_player_ratings
        )
        sum_participation_weight = sum(
            player.match_performance.projected_participation_weight
            for player in pre_match_player_ratings
        )

        return (
            team_rating / sum_participation_weight
            if sum_participation_weight > 0
            else 0
        )

    def _generate_pre_match_team_rating_value(
        self, pre_match_player_ratings: list[PreMatchPlayerRating]
    ) -> Optional[float]:
        if (
            len(pre_match_player_ratings) > 0
            and pre_match_player_ratings[0].match_performance.participation_weight
            is None
        ):
            return None
        team_rating = sum(
            player.rating_value * player.match_performance.participation_weight
            for player in pre_match_player_ratings
        )
        sum_participation_weight = sum(
            player.match_performance.participation_weight
            for player in pre_match_player_ratings
        )

        return (
            team_rating / sum_participation_weight
            if sum_participation_weight > 0
            else 0
        )

    def generate_rating_change(
        self,
        day_number: int,
        pre_match_team_rating: PreMatchTeamRating,
        pre_match_opponent_team_rating: PreMatchTeamRating,
    ) -> TeamRatingChange:

        player_rating_changes = []
        sum_participation_weight = 0
        sum_predicted_performance = 0
        sum_performance_value = 0
        sum_rating_change = 0

        for pre_player_rating in pre_match_team_rating.players:

            predicted_performance = self.performance_predictor.predict_performance(
                player_rating=pre_player_rating,
                opponent_team_rating=pre_match_opponent_team_rating,
                team_rating=pre_match_team_rating,
            )

            applied_rating_change_multiplier = (
                self._calculate_applied_rating_change_multiplier(
                    player_id=pre_player_rating.id, team_id=pre_match_team_rating.id
                )
            )
            performance_difference = (
                pre_player_rating.match_performance.performance_value
                - predicted_performance
            )
            rating_change_value = (
                performance_difference
                * applied_rating_change_multiplier
                * pre_player_rating.match_performance.participation_weight
            )
            if math.isnan(rating_change_value):
                raise ValueError(
                    f"rating_change_value is nan for {pre_player_rating.id}"
                )

            player_rating_change = PlayerRatingChange(
                id=pre_player_rating.id,
                predicted_performance=predicted_performance,
                participation_weight=pre_player_rating.match_performance.participation_weight,
                performance=pre_player_rating.match_performance.performance_value,
                rating_change_value=rating_change_value,
                league=pre_player_rating.league,
                day_number=day_number,
                pre_match_rating_value=pre_player_rating.rating_value,
            )
            player_rating_changes.append(player_rating_change)
            sum_predicted_performance += (
                player_rating_change.predicted_performance
                * pre_player_rating.match_performance.participation_weight
            )
            sum_performance_value += (
                pre_player_rating.match_performance.performance_value
                * pre_player_rating.match_performance.participation_weight
            )
            sum_rating_change += (
                player_rating_change.rating_change_value
                * pre_player_rating.match_performance.participation_weight
            )
            sum_participation_weight += (
                pre_player_rating.match_performance.participation_weight
            )

        predicted_performance = (
            sum_predicted_performance / sum_participation_weight
            if sum_participation_weight > 0
            else 0
        )
        performance = (
            sum_performance_value / sum_participation_weight
            if sum_participation_weight > 0
            else 0
        )

        return TeamRatingChange(
            players=player_rating_changes,
            id=pre_match_team_rating.id,
            rating_change_value=sum_rating_change,
            predicted_performance=predicted_performance,
            pre_match_projected_rating_value=pre_match_team_rating.rating_value,
            league=pre_match_team_rating.league,
            performance=performance,
        )

    def update_rating_by_team_rating_change(
        self,
        team_rating_change: TeamRatingChange,
        opponent_team_rating_change: TeamRatingChange,
    ) -> None:

        for player_rating_change in team_rating_change.players:
            id = player_rating_change.id
            self.player_ratings[id].confidence_sum = (
                self._calculate_post_match_confidence_sum(
                    entity_rating=self.player_ratings[id],
                    day_number=player_rating_change.day_number,
                    particpation_weight=player_rating_change.participation_weight,
                )
            )

            self.player_ratings[
                id
            ].rating_value += player_rating_change.rating_change_value
            self.player_ratings[
                id
            ].games_played += player_rating_change.participation_weight
            self.player_ratings[id].last_match_day_number = (
                player_rating_change.day_number
            )
            self.player_ratings[id].most_recent_team_id = team_rating_change.id

            self.start_rating_generator.update_players_to_leagues(
                rating_change=player_rating_change
            )
            if player_rating_change.league != opponent_team_rating_change.league:
                self._update_player_ratings_from_league_changes(
                    rating_change=player_rating_change
                )

    def _update_player_ratings_from_league_changes(
        self, rating_change: PlayerRatingChange
    ):

        league = rating_change.league

        if league not in self._league_rating_changes:
            self._league_rating_changes[rating_change.league] = 0
            self._league_rating_changes_count[league] = 0

        rating_change_value = rating_change.rating_change_value
        self._league_rating_changes[rating_change.league] += rating_change_value
        self._league_rating_changes_count[league] += 1

        if self._league_rating_changes[league] > abs(
            self.league_rating_change_update_threshold
        ):
            for player_id in self.start_rating_generator._league_to_player_ids[league]:
                self.player_ratings[player_id].rating_value += (
                    self._league_rating_changes[league]
                    * self.league_rating_adjustor_multiplier
                )

            self._league_rating_changes[league] = 0

    def _calculate_applied_rating_change_multiplier(
        self,
        player_id: str,
        team_id: str,
    ) -> float:
        if (
            self.player_ratings[player_id].most_recent_team_id
            and self.player_ratings[player_id].most_recent_team_id != team_id
        ):
            self.player_ratings[
                player_id
            ].confidence_sum -= self.team_id_change_confidence_sum_decrease

        min_applied_rating_change_multiplier = (
            self.rating_change_multiplier * self.min_rating_change_multiplier_ratio
        )
        confidence_change_multiplier = self.rating_change_multiplier * (
            (
                EXPECTED_MEAN_CONFIDENCE_SUM
                - self.player_ratings[player_id].confidence_sum
            )
            / self.confidence_value_denom
            + 1
        )
        applied_rating_change_multiplier = (
            confidence_change_multiplier * self.confidence_weight
            + (1 - self.confidence_weight) * self.rating_change_multiplier
        )

        # net_certain_sum_value = self.player_ratings[
        #                              player_id].confidence_sum - 20
        #  certain_factor = -(1 / (1 + math.exp(-net_certain_sum_value / 14)) - 0.5) * 2 + 1
        #   certain_multiplier = certain_factor * self.rating_change_multiplier
        #  multiplier = certain_multiplier * self.confidence_weight + (
        #          1 - self.confidence_weight) * self.rating_change_multiplier

        #   min_rating_change_multiplier = self.rating_change_multiplier * self.min_rating_change_multiplier_ratio
        return max(
            min_applied_rating_change_multiplier, applied_rating_change_multiplier
        )

    #     return max(min_applied_rating_change_multiplier, applied_rating_change_multiplier)

    def _calculate_post_match_confidence_sum(
        self, entity_rating: PlayerRating, day_number: int, particpation_weight: float
    ) -> float:
        days_ago = self._calculate_days_ago_since_last_match(
            last_match_day_number=entity_rating.last_match_day_number,
            day_number=day_number,
        )
        confidence_sum_value = (
            -min(days_ago, self.confidence_max_days)
            * self.confidence_days_ago_multiplier
            + entity_rating.confidence_sum
            + MATCH_CONTRIBUTION_TO_SUM_VALUE * particpation_weight
        )

        return max(0.0, min(confidence_sum_value, self.confidence_max_sum))

    def _calculate_days_ago_since_last_match(
        self, last_match_day_number, day_number: int
    ) -> float:
        match_day_number = day_number
        if last_match_day_number is None:
            return 0.0

        return match_day_number - last_match_day_number

    def _get_rating_by_id(self, id: str):
        if id not in self.player_ratings:
            raise KeyError(f"{id} not in player_ratings")
        return self.player_ratings[id]

    @property
    def teams(self) -> dict[str, Team]:
        return self._teams



from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from player_performance_ratings import ColumnNames

from player_performance_ratings.predictor._base import BasePredictor
from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformers.base_transformer import (
    BaseLagGenerator,
    BaseTransformer,
)


class CrossValidator(ABC):

    def __init__(self, scorer: Optional[BaseScorer]):
        self.scorer = scorer

    @abstractmethod
    def generate_validation_df(
        self,
        df: pd.DataFrame,
        column_names: ColumnNames,
        predictor: BasePredictor,
        estimator_features: list[str],
        return_features: bool,
        pre_lag_transformers: Optional[list[BaseTransformer]] = None,
        lag_generators: Optional[list[BaseLagGenerator]] = None,
        post_lag_transformers: Optional[list[BaseTransformer]] = None,
        add_train_prediction: bool = False,
    ) -> pd.DataFrame:
        pass

    def cross_validation_score(
        self, validation_df: pd.DataFrame, scorer: Optional[BaseScorer] = None
    ) -> float:
        if not scorer and not self.scorer:
            raise ValueError(
                "scorer is not defined. Either pass into constructor or as argument to method"
            )

        scorer = scorer or self.scorer
        return scorer.score(df=validation_df)



import copy
from typing import Optional
import polars as pl
import pandas as pd
from player_performance_ratings import ColumnNames

from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformers.base_transformer import (
    BaseTransformer,
    BaseLagGenerator,
)

from player_performance_ratings.cross_validator._base import CrossValidator
from player_performance_ratings.predictor._base import BasePredictor


class MatchKFoldCrossValidator(CrossValidator):
    def __init__(
        self,
        match_id_column_name: str,
        date_column_name: str,
        scorer: Optional[BaseScorer] = None,
        min_validation_date: Optional[str] = None,
        n_splits: int = 3,
    ):
        super().__init__(scorer=scorer)
        self.match_id_column_name = match_id_column_name
        self.date_column_name = date_column_name
        self.n_splits = n_splits
        self.min_validation_date = min_validation_date

    def generate_validation_df(
        self,
        df: pd.DataFrame,
        predictor: BasePredictor,
        column_names: ColumnNames,
        estimator_features: Optional[list[str]] = None,
        pre_lag_transformers: Optional[list[BaseTransformer]] = None,
        lag_generators: Optional[list[BaseLagGenerator]] = None,
        post_lag_transformers: Optional[list[BaseTransformer]] = None,
        return_features: bool = False,
        add_train_prediction: bool = False,
    ) -> pd.DataFrame:

        predictor = copy.deepcopy(predictor)
        validation_dfs = []
        ori_cols = df.columns.tolist()

        estimator_features = estimator_features or []
        pre_lag_transformers = pre_lag_transformers or []
        lag_generators = lag_generators or []
        post_lag_transformers = post_lag_transformers or []

        if not self.min_validation_date:
            unique_dates = df[self.date_column_name].unique()
            median_number = len(unique_dates) // 2
            self.min_validation_date = unique_dates[median_number]

        df = df.assign(__cv_match_number=range(len(df)))
        min_validation_match_number = df[
            df[self.date_column_name] >= self.min_validation_date
        ]["__cv_match_number"].min()
        if not pre_lag_transformers:
            for lag_transformer in lag_generators:
                lag_transformer.reset()
                df = lag_transformer.generate_historical(df, column_names=column_names)

        max_match_number = df["__cv_match_number"].max()
        train_cut_off_match_number = min_validation_match_number
        step_matches = (max_match_number - min_validation_match_number) / self.n_splits
        train_df = df[(df["__cv_match_number"] < train_cut_off_match_number)]
        if len(train_df) < 0:
            raise ValueError(
                f"train_df is empty. train_cut_off_day_number: {train_cut_off_match_number}. Select a lower validation_match value."
            )
        validation_df = df[
            (df["__cv_match_number"] >= train_cut_off_match_number)
            & (df["__cv_match_number"] <= train_cut_off_match_number + step_matches)
        ]

        for idx in range(self.n_splits):
            if pre_lag_transformers:
                for pre_lag_transformer in pre_lag_transformers:
                    pre_lag_transformer.reset()
                    train_df = pre_lag_transformer.fit_transform(
                        train_df, column_names=column_names
                    )
                    validation_df = pre_lag_transformer.transform(validation_df)
                for lag_transformer in lag_generators:
                    count_remaining_polars = [
                                                 l for l in lag_generators[idx:] if
                                                 "Polars" in l.__class__.__name__
                                             ] + [
                                                 l
                                                 for l in post_lag_transformers
                                                 if "Polars" in l.__class__.__name__
                                             ]

                    if isinstance(train_df, pd.DataFrame) and len(count_remaining_polars) == len(
                            train_df[idx:] + post_lag_transformers
                    ):
                        train_df = pl.from_pandas(train_df)


                    lag_transformer.reset()
                    train_df = lag_transformer.generate_historical(
                        train_df, column_names=column_names
                    )
                    validation_df = lag_transformer.generate_historical(
                        validation_df, column_names=column_names
                    )

            for post_lag_transformer in post_lag_transformers:
                post_lag_transformer.reset()
                train_df = post_lag_transformer.fit_transform(
                    train_df, column_names=column_names
                )
                validation_df = post_lag_transformer.transform(validation_df)

            predictor.train(train_df, estimator_features=estimator_features)

            if idx == 0 and add_train_prediction:
                train_df = train_df[
                    [c for c in train_df.columns if c not in predictor.columns_added]
                ]
                train_df = predictor.add_prediction(train_df)
                validation_dfs.append(train_df)

            validation_df = validation_df[
                [c for c in validation_df.columns if c not in predictor.columns_added]
            ]
            validation_df = predictor.add_prediction(validation_df)
            validation_dfs.append(validation_df)

            train_cut_off_match_number = train_cut_off_match_number + step_matches
            train_df = df[(df["__cv_match_number"] < train_cut_off_match_number)]

            if idx == self.n_splits - 2:
                validation_df = df[
                    (df["__cv_match_number"] >= train_cut_off_match_number)
                ]
            else:
                validation_df = df[
                    (df["__cv_match_number"] >= train_cut_off_match_number)
                    & (
                        df["__cv_match_number"]
                        < train_cut_off_match_number + step_matches
                    )
                ]

        concat_validation_df = pd.concat(validation_dfs).drop(
            columns=["__cv_match_number"]
        )
        if not return_features:
            concat_validation_df = concat_validation_df[
                ori_cols + predictor.columns_added
            ]

        return concat_validation_df.drop_duplicates(
            [column_names.match_id, column_names.team_id, column_names.player_id]
        )



import math
from dataclasses import dataclass
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Callable, Union, Any

import numpy as np
import pandas as pd

from player_performance_ratings.consts import PredictColumnNames


class Operator(Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUALS = ">="
    LESS_THAN_OR_EQUALS = "<="
    IN = "in"
    NOT_IN = "not in"


@dataclass
class Filter:
    column_name: str
    value: Union[Any, list[Any]]
    operator: Operator


def apply_filters(df: pd.DataFrame, filters: list[Filter]) -> pd.DataFrame:
    df = df.copy()
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df[df[filter.column_name] == filter.value]
        elif filter.operator == Operator.NOT_EQUALS:
            df = df[df[filter.column_name] != filter.value]
        elif filter.operator == Operator.GREATER_THAN:
            df = df[df[filter.column_name] > filter.value]
        elif filter.operator == Operator.LESS_THAN:
            df = df[df[filter.column_name] < filter.value]
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df[df[filter.column_name] >= filter.value]
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df[df[filter.column_name] <= filter.value]
        elif filter.operator == Operator.IN:
            df = df[df[filter.column_name].isin(filter.value)]
        elif filter.operator == Operator.NOT_IN:
            df = df[~df[filter.column_name].isin(filter.value)]

    return df


class BaseScorer(ABC):

    def __init__(
        self,
        target: str,
        pred_column: str,
        filters: Optional[list[Filter]] = None,
        granularity: Optional[list[str]] = None,
    ):
        self.target = target
        self.pred_column = pred_column
        self.filters = filters or []
        self.granularity = granularity

    @abstractmethod
    def score(self, df: pd.DataFrame) -> float:
        pass


class SklearnScorer(BaseScorer):

    def __init__(
        self,
        pred_column: str,
        scorer_function: Callable,
        target: Optional[str] = PredictColumnNames.TARGET,
        granularity: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
    ):
        self.pred_column_name = pred_column
        self.scorer_function = scorer_function
        super().__init__(
            target=target,
            pred_column=pred_column,
            granularity=granularity,
            filters=filters,
        )

    def score(self, df: pd.DataFrame) -> float:
        df = df.copy()
        df = apply_filters(df, self.filters)
        if self.granularity:
            grouped = (
                df.groupby(self.granularity)[self.pred_column_name, self.target]
                .mean()
                .reset_index()
            )
        else:
            grouped = df
        if isinstance(df[self.pred_column_name].iloc[0], list):
            return self.scorer_function(
                grouped[self.target],
                np.asarray(grouped[self.pred_column_name]).tolist(),
            )
        return self.scorer_function(
            grouped[self.target], grouped[self.pred_column_name]
        )


class OrdinalLossScorer(BaseScorer):

    def __init__(
        self,
        pred_column: str,
        targets_to_measure: Optional[list[int]] = None,
        target: Optional[str] = PredictColumnNames.TARGET,
        granularity: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
    ):

        self.pred_column_name = pred_column
        self.targets_to_measure = targets_to_measure
        self.granularity = granularity
        super().__init__(
            target=target,
            pred_column=pred_column,
            filters=filters,
            granularity=granularity,
        )

    def score(self, df: pd.DataFrame) -> float:
        df = df.copy()
        df = apply_filters(df, self.filters)
        df.reset_index(drop=True, inplace=True)

        distinct_classes_variations = df.drop_duplicates(subset=["classes"])[
            "classes"
        ].tolist()

        sum_lrs = [0 for _ in range(len(distinct_classes_variations))]
        sum_lr = 0
        for variation_idx, distinct_class_variation in enumerate(
            distinct_classes_variations
        ):

            if not isinstance(distinct_class_variation, list):
                if math.isnan(distinct_class_variation):
                    continue

            rows_target_group = df[
                df["classes"].apply(lambda x: x == distinct_class_variation)
            ]
            probs = rows_target_group[self.pred_column_name]
            last_column_name = f"prob_under_{distinct_class_variation[0] - 0.5}"
            rows_target_group[last_column_name] = probs.apply(lambda x: x[0])

            for idx, class_ in enumerate(distinct_class_variation[1:]):

                p_c = "prob_under_" + str(class_ + 0.5)
                rows_target_group[p_c] = (
                    probs.apply(lambda x: x[idx + 1])
                    + rows_target_group[last_column_name]
                )

                count_exact = len(
                    rows_target_group[rows_target_group["__target"] == class_]
                )
                weight_class = count_exact / len(rows_target_group)

                if self.granularity:
                    grouped = (
                        rows_target_group.groupby(self.granularity + ["__target"])[p_c]
                        .mean()
                        .reset_index()
                    )
                else:
                    grouped = rows_target_group

                grouped["min"] = 0.0001
                grouped["max"] = 0.9999
                grouped[p_c] = np.minimum(grouped["max"], grouped[p_c])
                grouped[p_c] = np.maximum(grouped["min"], grouped[p_c])
                grouped["log_loss"] = 0
                grouped.loc[grouped["__target"] <= class_, "log_loss"] = np.log(
                    grouped[p_c]
                )
                grouped.loc[grouped["__target"] > class_, "log_loss"] = np.log(
                    1 - grouped[p_c]
                )
                log_loss = grouped["log_loss"].mean()
                sum_lrs[variation_idx] -= log_loss * weight_class

                last_column_name = p_c
            sum_lr += sum_lrs[variation_idx] * len(rows_target_group) / len(df)
        return sum_lr



from typing import Optional, Union

import numpy as np
import pandas as pd
import polars as pl

from player_performance_ratings import ColumnNames
from player_performance_ratings.transformers.base_transformer import (
    BaseLagGenerator,
    BaseLagGeneratorPolars,
)
from player_performance_ratings.utils import validate_sorting


def create_output_column_by_game_group(
    data: pd.DataFrame,
    feature_name: str,
    weight_column: str,
    game_id: str,
    granularity: list[str],
) -> pd.DataFrame:
    if weight_column:
        data = data.assign(
            **{f"weighted_{feature_name}": data[feature_name] * data[weight_column]}
        )
        data = data.assign(
            **{
                "sum_weighted": data.groupby(granularity + [game_id])[
                    weight_column
                ].transform("sum")
            }
        )
        data = data.assign(
            **{feature_name: data[f"weighted_{feature_name}"] / data["sum_weighted"]}
        )
        data = data.groupby(granularity + [game_id])[feature_name].sum().reset_index()

    else:
        data = data.groupby(granularity + [game_id])[feature_name].mean().reset_index()
    return data


class LagTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        lag_length: int,
        granularity: Optional[list[str]] = None,
        days_between_lags: Optional[list[int]] = None,
        future_lag: bool = False,
        prefix: str = "lag_",
        add_opponent: bool = False,
    ):
        """
        :param features. List of features to create lags for
        :param lag_length: Number of lags
        :param granularity: Columns to group by before lagging. E.g. player_id or [player_id, position].
            In the latter case it will get the lag for each player_id and position combination.
            Defaults to
        :param days_between_lags. Adds a column with the number of days between the lagged date and the current date.
        :param future_lag: If True, the lag will be calculated for the future instead of the past.
        :param prefix:
            Prefix for the new lag columns
        """

        super().__init__(
            features=features,
            add_opponent=add_opponent,
            prefix=prefix,
            iterations=[i for i in range(1, lag_length + 1)],
            granularity=granularity,
        )
        self.days_between_lags = days_between_lags or []
        for days_lag in self.days_between_lags:
            self._features_out.append(f"{prefix}{days_lag}_days_ago")

        self.lag_length = lag_length
        self.future_lag = future_lag
        self._df = None

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        """ """
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        if len(
            df.drop_duplicates(
                subset=[
                    self.column_names.player_id,
                    self.column_names.team_id,
                    self.column_names.match_id,
                ]
            )
        ) != len(df):
            raise ValueError(
                f"Duplicated rows in df. Df must be a unique combination of {self.column_names.player_id} and {self.column_names.update_match_id}"
            )

        concat_df = self._concat_df(df=df)

        if self.column_names.participation_weight:
            for feature in self.features:
                concat_df = concat_df.assign(
                    **{
                        feature: concat_df[feature]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

        grouped = (
            concat_df.groupby(
                self.granularity
                + [self.column_names.update_match_id, self.column_names.start_date]
            )[self.features]
            .mean()
            .reset_index()
        )

        grouped = grouped.sort_values(
            [self.column_names.start_date, self.column_names.update_match_id]
        )

        for days_lag in self.days_between_lags:
            if self.future_lag:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[
                    self.column_names.start_date
                ].shift(-days_lag)
                grouped[f"{self.prefix}{days_lag}_days_ago"] = (
                    pd.to_datetime(grouped["shifted_days"])
                    - pd.to_datetime(grouped[self.column_names.start_date])
                ).dt.days
            else:
                grouped["shifted_days"] = grouped.groupby(self.granularity)[
                    self.column_names.start_date
                ].shift(days_lag)
                grouped[f"{self.prefix}{days_lag}_days_ago"] = (
                    pd.to_datetime(grouped[self.column_names.start_date])
                    - pd.to_datetime(grouped["shifted_days"])
                ).dt.days

            grouped = grouped.drop(columns=["shifted_days"])

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f"{self.prefix}{lag}_{feature_name}"
                if output_column_name in concat_df.columns:
                    raise ValueError(
                        f"Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed"
                    )

        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                output_column_name = f"{self.prefix}{lag}_{feature_name}"
                if self.future_lag:
                    grouped = grouped.assign(
                        **{
                            output_column_name: grouped.groupby(self.granularity)[
                                feature_name
                            ].shift(-lag)
                        }
                    )
                else:
                    grouped = grouped.assign(
                        **{
                            output_column_name: grouped.groupby(self.granularity)[
                                feature_name
                            ].shift(lag)
                        }
                    )

        feats_out = []
        for feature_name in self.features:
            for lag in range(1, self.lag_length + 1):
                feats_out.append(f"{self.prefix}{lag}_{feature_name}")

        for days_lag in self.days_between_lags:
            feats_out.append(f"{self.prefix}{days_lag}_days_ago")

        concat_df = concat_df.merge(
            grouped[
                self.granularity
                + [
                    self.column_names.update_match_id,
                    self.column_names.start_date,
                    *feats_out,
                ]
            ],
            on=self.granularity
            + [self.column_names.update_match_id, self.column_names.start_date],
            how="left",
        )

        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        window: int,
        granularity: Union[list[str], str] = None,
        add_opponent: bool = False,
        min_periods: int = 1,
        are_estimator_features=True,
        prefix: str = "rolling_mean_",
    ):
        """

        :param features:
            Features to create rolling mean for

        :param granularity:
            Columns to group by before rolling mean. E.g. player_id or [player_id, position].
             In the latter case it will get the rolling mean for each player_id and position combination.

        :param window:
            Window size for rolling mean, if 10 will calculate rolling mean over the prior 10 observations

        :param game_id:
            Column name of game_id.
            If there are more multiple rows per granularity per game_id and you want to add the rolling mean per game_id, set game_id
            This will calculate the mean of the features per game and the rolling mean will be calculated on that.

        :param weight_column:
            Only used if game_id is set.
            Will calculate weighted mean of the features per game and the rolling mean will be the prior weighted means.
            This is useful when working with partial game-data of different lengths.
            In that case it can beneficial to set the game-length as weight_colum.

        :param df: Optional parameter to pass in a dataframe to calculate the rolling mean on.
            If not passed in, it will use the dataframe passed in the transform method.
            This is useful if you want to calculate the rolling mean on a different dataframe than the one you want to transform.
            Will merge the two dataframes on game_id and granularity.

        :param prefix:
            Prefix for the new rolling mean columns
        """
        super().__init__(
            features=features,
            add_opponent=add_opponent,
            iterations=[window],
            prefix=prefix,
            granularity=granularity,
            are_estimator_features=are_estimator_features,
        )
        self.window = window
        self.min_periods = min_periods

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)

        for feature_name in self.features:

            if self.column_names.participation_weight:
                concat_df = concat_df.assign(
                    **{
                        feature_name: concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

            output_column_name = f"{self.prefix}{self.window}_{feature_name}"
            if output_column_name in concat_df.columns:
                raise ValueError(
                    f"Column {output_column_name} already exists. Choose different prefix or ensure no duplication was performed"
                )

            agg_dict = {feature_name: "mean", self.column_names.start_date: "first"}
            grp = (
                concat_df.groupby(
                    self.granularity + [self.column_names.update_match_id]
                )
                .agg(agg_dict)
                .reset_index()
            )
            grp.sort_values(
                by=[self.column_names.start_date, self.column_names.update_match_id],
                inplace=True,
            )

            grp = grp.assign(
                **{
                    output_column_name: grp.groupby(self.granularity)[
                        feature_name
                    ].transform(
                        lambda x: x.shift()
                        .rolling(self.window, min_periods=self.min_periods)
                        .mean()
                    )
                }
            )

            concat_df = concat_df.merge(
                grp[
                    self.granularity
                    + [self.column_names.update_match_id, output_column_name]
                ],
                on=self.granularity + [self.column_names.update_match_id],
                how="left",
            )
            concat_df = concat_df.sort_values(
                by=[
                    self.column_names.start_date,
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                ]
            )

        feats_added = [f for f in self.features_out if f in concat_df.columns]

        concat_df[feats_added] = concat_df.groupby(self.granularity)[
            feats_added
        ].fillna(method="ffill")
        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class RollingMeanDaysTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        days: Union[int, list[int]],
        granularity: Union[list[str], str] = None,
        add_count: bool = False,
        add_opponent: bool = False,
        prefix: str = "rolling_mean_days_",
    ):
        self.days = days
        if isinstance(self.days, int):
            self.days = [self.days]
        super().__init__(
            features=features,
            iterations=[i for i in self.days],
            prefix=prefix,
            add_opponent=add_opponent,
            granularity=granularity,
        )

        self.add_count = add_count
        self._fitted_game_ids = []

        for day in self.days:
            if self.add_count:
                feature = f"{self.prefix}{day}_count"
                self._features_out.append(feature)
                self._entity_features.append(feature)

                if self.add_opponent:
                    self._features_out.append(f"{self.prefix}{day}_count_opponent")

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        ori_types = {c: df[c].dtype for c in df.columns}
        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        df = self._create_transformed_df(df=df, concat_df=concat_df)

        if self.add_count:
            for day in self.days:
                df = df.assign(
                    **{
                        f"{self.prefix}{day}_count": df[
                            f"{self.prefix}{day}_count"
                        ].fillna(0)
                    }
                )
                if self.add_opponent:
                    df = df.assign(
                        **{
                            f"{self.prefix}{day}_count_opponent": df[
                                f"{self.prefix}{day}_count_opponent"
                            ].fillna(0)
                        }
                    )
        if "is_future" in df.columns:
            df = df.drop(columns="is_future")

        for c, t in ori_types.items():
            if c not in self.features_out and df[c].dtype != t:
                df[c] = df[c].astype(t)
        return df

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        ori_types = {c: df[c].dtype for c in df.columns}
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")

        for c, t in ori_types.items():
            if c not in self.features_out and transformed_future[c].dtype != t:
                transformed_future[c] = transformed_future[c].astype(t)
        return transformed_future

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)

        for feature_name in self.features:

            if self.column_names.participation_weight:
                concat_df = concat_df.assign(
                    **{
                        feature_name: concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    }
                )

        concat_df = concat_df.assign(
            **{"__date_day": lambda x: x[self.column_names.start_date].dt.date}
        )

        aggregations = {feature: "mean" for feature in self.features}
        aggregations["__date_day"] = "first"
        grouped = (
            concat_df.groupby([*self.granularity, self.column_names.update_match_id])
            .agg(aggregations)
            .reset_index()
        )
        count_feats = []
        for day in self.days:
            prefix_day = f"{self.prefix}{day}"
            if self.add_count:
                count_feats.append(f"{prefix_day}_count")
            grouped = self._add_rolling_feature(
                concat_df=grouped,
                day=day,
                granularity=self.granularity,
                prefix_day=prefix_day,
            )

        feats_created = [f for f in self.features_out if f in grouped.columns]
        concat_df = concat_df.merge(
            grouped[
                [*self.granularity, *feats_created, self.column_names.update_match_id]
            ],
            on=self.granularity + [self.column_names.update_match_id],
            how="left",
        )
        concat_df = concat_df.sort_values(
            by=[
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )
        return concat_df.drop(columns="__date_day")

    def _add_rolling_feature(
        self, concat_df: pd.DataFrame, day: int, granularity: list[str], prefix_day: str
    ):

        if len(granularity) > 1:
            granularity_concat = "__".join(granularity)
            temporary_str_df = concat_df[granularity].astype(str)
            concat_df[granularity_concat] = temporary_str_df.agg("__".join, axis=1)
        else:
            granularity_concat = granularity[0]

        concat_df["is_nan"] = concat_df[self.features[0]].isna().astype(float)

        df1 = (
            concat_df.groupby(["__date_day", granularity_concat])[
                [*self.features, "is_nan"]
            ]
            .agg(["sum", "size"])
            .unstack()
            .asfreq("d", fill_value=np.nan)
            .rolling(window=day, min_periods=1)
            .sum()
            .shift()
            .stack()
        )
        feats = []
        for feature_name in self.features:
            feats.append(f"{prefix_day}_{feature_name}_sum")
            feats.append(f"{prefix_day}_{feature_name}_count")

        df1.columns = feats + ["is_nan_sum", "is_nan_count"]
        for feature_name in self.features:
            df1[f"{prefix_day}_{feature_name}_count"] = (
                df1[f"{prefix_day}_{feature_name}_count"] - df1["is_nan_sum"]
            )
            df1[f"{prefix_day}_{feature_name}"] = df1[
                f"{prefix_day}_{feature_name}_sum"
            ] / (df1[f"{prefix_day}_{feature_name}_count"])

            df1.loc[
                df1[f"{prefix_day}_{feature_name}_count"] == 0,
                f"{prefix_day}_{feature_name}",
            ] = np.nan

        if self.add_count:
            df1[f"{prefix_day}_count"] = df1[f"{prefix_day}_{self.features[0]}_count"]
            df1 = df1.drop(
                columns=[
                    f"{prefix_day}_{feature_name}_count"
                    for feature_name in self.features
                ]
            )

        concat_df["__date_day"] = pd.to_datetime(concat_df["__date_day"])
        concat_df = concat_df.merge(
            df1[[c for c in df1.columns if c in self.features_out]].reset_index(),
            on=["__date_day", granularity_concat],
        )

        return concat_df

    def reset(self):
        self._df = None
        self._fitted_game_ids = []

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class BinaryOutcomeRollingMeanTransformer(BaseLagGenerator):

    def __init__(
        self,
        features: list[str],
        window: int,
        binary_column: str,
        granularity: list[str] = None,
        prob_column: Optional[str] = None,
        min_periods: int = 1,
        add_opponent: bool = False,
        prefix: str = "rolling_mean_binary_",
    ):
        super().__init__(
            features=features,
            add_opponent=add_opponent,
            prefix=prefix,
            iterations=[],
            granularity=granularity,
        )
        self.window = window
        self.min_periods = min_periods
        self.binary_column = binary_column
        self.prob_column = prob_column
        for feature_name in self.features:
            feature1 = f"{self.prefix}{self.window}_{feature_name}_1"
            feature2 = f"{self.prefix}{self.window}_{feature_name}_0"
            self._features_out.append(f"{self.prefix}{self.window}_{feature_name}_1")
            self._features_out.append(f"{self.prefix}{self.window}_{feature_name}_0")
            self._entity_features.append(feature1)
            self._entity_features.append(feature2)

            if self.add_opponent:
                self._features_out.append(
                    f"{self.prefix}{self.window}_{feature_name}_1_opponent"
                )
                self._features_out.append(
                    f"{self.prefix}{self.window}_{feature_name}_0_opponent"
                )

        if self.prob_column:
            for feature_name in self.features:
                prob_feature = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                self._features_out.append(prob_feature)

        self._estimator_features_out = self._features_out.copy()

    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        if df[self.binary_column].dtype in ("float64", "float32", "float16", "float"):
            df = df.assign(**{self.binary_column: df[self.binary_column].astype(int)})

        df = df.assign(is_future=0)
        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        validate_sorting(df=df, column_names=self.column_names)
        additional_cols_to_use = [self.binary_column] + (
            [self.prob_column] if self.prob_column else []
        )
        self._store_df(df, additional_cols_to_use=additional_cols_to_use)
        concat_df = self._generate_concat_df_with_feats(df)
        concat_df = self._add_weighted_prob(transformed_df=concat_df)
        transformed_df = self._create_transformed_df(df=df, concat_df=concat_df)
        if "is_future" in transformed_df.columns:
            transformed_df = transformed_df.drop(columns="is_future")
        return self._add_weighted_prob(transformed_df=transformed_df)

    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._df is None:
            raise ValueError(
                "generate_historical needs to be called before generate_future"
            )

        if self.binary_column in df.columns:
            if df[self.binary_column].dtype in (
                "float64",
                "float32",
                "float16",
                "float",
            ):
                df = df.assign(
                    **{self.binary_column: df[self.binary_column].astype(int)}
                )
        df = df.assign(is_future=1)
        concat_df = self._generate_concat_df_with_feats(df=df)
        transformed_df = concat_df[
            concat_df[self.column_names.match_id].isin(
                df[self.column_names.match_id].astype("str").unique().tolist()
            )
        ]
        transformed_future = self._generate_future_feats(
            transformed_df=transformed_df,
            ori_df=df,
            known_future_features=self._get_known_future_features(),
        )
        if "is_future" in transformed_future.columns:
            transformed_future = transformed_future.drop(columns="is_future")
        return self._add_weighted_prob(transformed_df=transformed_future)

    def _get_known_future_features(self) -> list[str]:
        known_future_features = []
        if self.prob_column:
            for idx, feature_name in enumerate(self.features):
                weighted_prob_feat_name = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                known_future_features.append(weighted_prob_feat_name)

        return known_future_features

    def _add_weighted_prob(self, transformed_df: pd.DataFrame) -> pd.DataFrame:

        if self.prob_column:
            for idx, feature_name in enumerate(self.features):
                weighted_prob_feat_name = (
                    f"{self.prefix}{self.window}_{self.prob_column}_{feature_name}"
                )
                transformed_df[weighted_prob_feat_name] = transformed_df[
                    f"{self.prefix}{self.window}_{feature_name}_1"
                ] * transformed_df[self.prob_column] + transformed_df[
                    f"{self.prefix}{self.window}_{feature_name}_0"
                ] * (
                    1 - transformed_df[self.prob_column]
                )
        return transformed_df

    def _generate_concat_df_with_feats(self, df: pd.DataFrame) -> pd.DataFrame:

        additional_cols_to_use = [self.binary_column] + (
            [self.prob_column] if self.prob_column else []
        )
        concat_df = self._concat_df(df, additional_cols_to_use=additional_cols_to_use)
        aggregation = {
            **{f: "mean" for f in self.features},
            **{self.binary_column: "first"},
        }
        grouped = (
            concat_df.groupby(
                [
                    self.column_names.update_match_id,
                    *self.granularity,
                    self.column_names.start_date,
                    "is_future",
                ]
            )
            .agg(aggregation)
            .reset_index()
        )

        grouped = grouped.sort_values(
            by=[
                self.column_names.start_date,
                "is_future",
                self.column_names.update_match_id,
            ]
        )

        feats_added = []
        for feature in self.features:
            mask_result_1 = grouped[self.binary_column] == 1
            mask_result_0 = grouped[self.binary_column] == 0

            grouped["value_result_1"] = grouped[feature].where(mask_result_1)
            grouped["value_result_0"] = grouped[feature].where(mask_result_0)

            grouped[f"{self.prefix}{self.window}_{feature}_1"] = grouped.groupby(
                self.granularity
            )["value_result_1"].transform(
                lambda x: x.shift().rolling(window=self.window, min_periods=1).mean()
            )
            grouped[f"{self.prefix}{self.window}_{feature}_0"] = grouped.groupby(
                self.granularity
            )["value_result_0"].transform(
                lambda x: x.shift().rolling(window=self.window, min_periods=1).mean()
            )

            feats_added.append(f"{self.prefix}{self.window}_{feature}_1")
            feats_added.append(f"{self.prefix}{self.window}_{feature}_0")

        grouped["count_result_1"] = (
            grouped[grouped[self.binary_column] == 1]
            .groupby([*self.granularity])
            .cumcount()
            + 1
        )
        grouped["count_result_0"] = (
            grouped[grouped[self.binary_column] == 0]
            .groupby([*self.granularity])
            .cumcount()
            + 1
        )
        grouped[["count_result_0", "count_result_1"]] = grouped.groupby(
            self.granularity
        )[["count_result_0", "count_result_1"]].fillna(method="ffill")
        for feature in self.features:
            grouped.loc[
                grouped["count_result_1"] < self.min_periods,
                f"{self.prefix}{self.window}_{feature}_1",
            ] = np.nan
            grouped.loc[
                grouped["count_result_0"] < self.min_periods,
                f"{self.prefix}{self.window}_{feature}_0",
            ] = np.nan

        concat_df = concat_df.merge(
            grouped[
                [self.column_names.update_match_id, *self.granularity, *feats_added]
            ],
            on=[self.column_names.update_match_id, *self.granularity],
            how="left",
        )

        concat_df[feats_added] = concat_df.groupby(self.granularity)[
            feats_added
        ].fillna(method="ffill")
        return concat_df


class RollingMeanTransformerPolars(BaseLagGeneratorPolars):

    def __init__(
        self,
        features: list[str],
        window: int,
        granularity: Union[list[str], str] = None,
        add_opponent: bool = False,
        min_periods: int = 1,
        are_estimator_features=True,
        prefix: str = "rolling_mean_",
    ):

        super().__init__(
            features=features,
            add_opponent=add_opponent,
            iterations=[window],
            prefix=prefix,
            granularity=granularity,
            are_estimator_features=are_estimator_features,
        )
        self.window = window
        self.min_periods = min_periods

    def generate_historical(
        self, df: Union[pd.DataFrame, pl.DataFrame], column_names: ColumnNames
    ) -> Union[pl.DataFrame, pd.DataFrame]:
        if isinstance(df, pd.DataFrame):
            ori_type = "pd"
            df = pl.DataFrame(df)
        else:
            ori_type = "pl"
        if (
            df.unique(
                subset=[
                    column_names.player_id,
                    column_names.team_id,
                    column_names.match_id,
                ]
            ).shape[0]
            != df.shape[0]
        ):
            raise ValueError(
                f"Duplicated rows in df. Df must be a unique combination of {column_names.player_id} and {column_names.update_match_id}"
            )

        df = df.with_columns(pl.lit(0).alias("is_future"))

        self.column_names = column_names
        self.granularity = self.granularity or [self.column_names.player_id]
        self._store_df(df)
        concat_df = self._generate_concat_df_with_feats(df)
        transformed_df = self._create_transformed_df(df=df, concat_df=concat_df)
        cn = self.column_names
        recasts_mapping = {}
        for c in [cn.player_id, cn.team_id, cn.match_id]:
            if transformed_df[c].dtype != df[c].dtype:
                recasts_mapping[c] = df[c].dtype
        transformed_df = transformed_df.cast(recasts_mapping)

        df = df.join(
            transformed_df.select(
                cn.player_id, cn.team_id, cn.match_id, *self.features_out
            ),
            on=[cn.player_id, cn.team_id, cn.match_id],
            how="left",
        )
        if "is_future" in df.columns:
            df = df.drop("is_future")
        if ori_type == "pd":
            return df.to_pandas()
        return df

    def generate_future(
        self, df: Union[pd.DataFrame, pl.DataFrame]
    ) -> Union[pl.DataFrame, pd.DataFrame]:
        if isinstance(df, pd.DataFrame):
            ori_type = "pd"
            df = pl.DataFrame(df)
        else:
            ori_type = "pl"

        df = df.with_columns(pl.lit(1).alias("is_future"))
        concat_df = self._generate_concat_df_with_feats(df=df)
        unique_match_ids = df.select(
            pl.col(self.column_names.match_id).cast(pl.Utf8).unique()
        ).to_series()
        transformed_df = concat_df.filter(
            pl.col(self.column_names.match_id).cast(pl.Utf8).is_in(unique_match_ids)
        )
        transformed_df = self._generate_future_feats(
            transformed_df=transformed_df, ori_df=df
        )
        cn = self.column_names
        recasts_mapping = {}
        for c in [cn.player_id, cn.team_id, cn.match_id]:
            if transformed_df[c].dtype != df[c].dtype:
                recasts_mapping[c] = df[c].dtype
        transformed_df = transformed_df.cast(recasts_mapping)

        df = df.join(
            transformed_df.select(
                cn.player_id, cn.team_id, cn.match_id, *self.features_out
            ),
            on=[cn.player_id, cn.team_id, cn.match_id],
            how="left",
        )
        if "is_future" in df.columns:
            df = df.drop("is_future")
        if ori_type == "pd":
            return df.to_pandas()
        return df

    def _generate_concat_df_with_feats(self, df: pl.DataFrame) -> pl.DataFrame:

        if self._df is None:
            raise ValueError("fit_transform needs to be called before transform")

        concat_df = self._concat_df(df)
        if self.column_names.participation_weight:
            concat_df = concat_df.with_columns(
                [
                    (
                        concat_df[feature_name]
                        * concat_df[self.column_names.participation_weight]
                    ).alias(feature_name)
                    for feature_name in self.features
                ]
            )

        agg_dict = [
            pl.col(feature_name).mean().alias(feature_name)
            for feature_name in self.features
        ]
        agg_dict.append(
            pl.col(self.column_names.start_date)
            .first()
            .alias(self.column_names.start_date)
        )

        grp = concat_df.group_by(
            self.granularity + [self.column_names.update_match_id]
        ).agg(agg_dict)
        grp = grp.filter(pl.col(self.column_names.start_date).is_not_null())
        grp = grp.sort(
            [self.column_names.start_date, self.column_names.update_match_id]
        )

        rolling_means = [
            pl.col(feature_name)
            .shift()
            .rolling_mean(window_size=self.window, min_periods=self.min_periods)
            .over(self.granularity)
            .alias(f"{self.prefix}{self.window}_{feature_name}")
            for feature_name in self.features
        ]
        grp = grp.with_columns(rolling_means)

        selection_columns = (
            self.granularity
            + [self.column_names.update_match_id]
            + [f"{self.prefix}{self.window}_{feature}" for feature in self.features]
        )
        concat_df = concat_df.join(
            grp.select(selection_columns),
            on=self.granularity + [self.column_names.update_match_id],
            how="left",
        )

        concat_df = concat_df.sort(
            [
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )

        feats_added = [f for f in self.features_out if f in concat_df.columns]

        concat_df = concat_df.with_columns(
            [
                pl.col(f).forward_fill().over(self.granularity).alias(f)
                for f in feats_added
            ]
        )
        return concat_df

    @property
    def features_out(self) -> list[str]:
        return self._features_out



import logging
from abc import abstractmethod, ABC
from typing import Optional

import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder, StandardScaler

from player_performance_ratings.predictor_transformer import (
    PredictorTransformer,
    SkLearnTransformerWrapper,
    ConvertDataFrameToCategoricalTransformer,
)


class BasePredictor(ABC):

    def __init__(
        self,
        estimator,
        estimator_features: list[str],
        target: str,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
        pred_column: Optional[str] = None,
        filters: Optional[dict] = None,
        auto_pre_transform: bool = True,
    ):
        self._estimator_features = estimator_features or []
        self.estimator = estimator
        self._target = target
        self._pred_column = pred_column or f"{self._target}_prediction"
        self.pre_transformers = pre_transformers or []
        self._deepest_estimator = self.estimator
        self.filters = filters or []
        self.multiclassifier = False
        self._classes_ = None
        self.auto_pre_transform = auto_pre_transform

        iterations = 0
        while hasattr(self._deepest_estimator, "estimator"):
            self._deepest_estimator = self._deepest_estimator.estimator
            iterations += 1
            if iterations > 10:
                raise ValueError("estimator is too deep")

    @property
    def estimator_type(self) -> str:
        if hasattr(self._deepest_estimator, "predict_proba"):
            return "classifier"
        return "regressor"

    @property
    def deepest_estimator(self) -> object:
        return self._deepest_estimator

    @abstractmethod
    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        pass

    @abstractmethod
    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def pred_column(self) -> str:
        return self._pred_column

    @property
    def target(self) -> str:
        return self._target

    @property
    def columns_added(self) -> list[str]:
        if not self.multiclassifier:
            return [self.pred_column]
        return [self.pred_column, "classes"]

    @property
    def classes_(self) -> Optional[list[str]]:
        if self._classes_:
            return self._classes_
        if "classes_" not in dir(self.estimator):
            return None
        return self.estimator.classes_

    def set_target(self, new_target_name: str):
        self._target = new_target_name

    def _create_pre_transformers(self, df: pd.DataFrame) -> list[PredictorTransformer]:
        pre_transformers = []
        feats_to_transform = []
        for estimator_feature in self._estimator_features.copy():

            if estimator_feature not in df.columns:
                self._estimator_features.remove(estimator_feature)
                logging.warning(
                    f"Feature {estimator_feature} not in df, removing from estimator_features"
                )

                feats_to_transform.append(estimator_feature)

        if feats_to_transform:
            if self._deepest_estimator.__class__.__name__ in (
                "LogisticRegression",
                "LinearRegression",
            ):
                logging.info(
                    f"Adding OneHotEncoder to pre_transformers for features: {feats_to_transform}"
                )
                pre_transformers.append(
                    SkLearnTransformerWrapper(
                        transformer=OneHotEncoder(handle_unknown="ignore"),
                        features=feats_to_transform,
                    )
                )

            elif self._deepest_estimator.__class__.__name__ in (
                "LGBMRegressor",
                "LGBMClassifier",
            ):
                logging.info(
                    f"Adding ConvertDataFrameToCategoricalTransformer to pre_transformers for features: {feats_to_transform}"
                )
                pre_transformers.append(
                    ConvertDataFrameToCategoricalTransformer(
                        features=feats_to_transform
                    )
                )

        for pre_transformer in self.pre_transformers:
            for feature in pre_transformer.features:
                if feature in self._estimator_features:
                    self._estimator_features.remove(feature)

        if self._deepest_estimator.__class__.__name__ in (
            "LogisticRegression",
            "LinearRegression",
        ):
            if "StandardScaler" not in [
                pre_transformer.transformer.__class__.__name__
                for pre_transformer in self.pre_transformers
                if hasattr(pre_transformer, "transformer")
            ]:
                logging.info(f"Adding StandardScaler to pre_transformers")
                pre_transformers.append(
                    SkLearnTransformerWrapper(
                        transformer=StandardScaler(),
                        features=self._estimator_features.copy(),
                    )
                )

            if "SimpleImputer" not in [
                pre_transformer.transformer.__class__.__name__
                for pre_transformer in self.pre_transformers
                if hasattr(pre_transformer, "transformer")
            ]:
                pre_transformers.append(
                    SkLearnTransformerWrapper(
                        transformer=SimpleImputer(),
                        features=self._estimator_features.copy(),
                    )
                )
        return pre_transformers

    def fit_transform_pre_transformers(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.auto_pre_transform:
            self.pre_transformers += self._create_pre_transformers(df)


        for pre_transformer in self.pre_transformers:
            values = pre_transformer.fit_transform(df)
            features_out = pre_transformer.features_out
            df[features_out] = values
            self._estimator_features = list(
                set(pre_transformer.features_out + self._estimator_features)
            )

        return df

    def transform_pre_transformers(self, df: pd.DataFrame) -> pd.DataFrame:
        for pre_transformer in self.pre_transformers:
            values = pre_transformer.transform(df)
            features_out = pre_transformer.features_out
            df[features_out] = values
        return df

    @property
    def estimator_features(self) -> list[str]:
        return self._estimator_features



import pickle

import pandas as pd
from sklearn.linear_model import LogisticRegression

from player_performance_ratings.pipeline import Pipeline
from player_performance_ratings.predictor import GameTeamPredictor
from player_performance_ratings.tuner.performances_generator_tuner import PerformancesSearchRange
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner

from player_performance_ratings.tuner.rating_generator_tuner import UpdateRatingGeneratorTuner
from player_performance_ratings.ratings import UpdateRatingGenerator

from player_performance_ratings.data_structures import ColumnNames

from player_performance_ratings.tuner import PipelineTuner, PerformancesGeneratorTuner
from player_performance_ratings.tuner.utils import ParameterSearchRange, get_default_team_rating_search_range

column_names = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    league='league',
    position='position',
)
df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])
df['champion_position'] = df['champion'] + df['position']
df['__target'] = df['result']


df = df.drop_duplicates(subset=['gameid', 'teamname', 'playername'])

df = (
    df.assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)
df = df.drop_duplicates(subset=['gameid', 'teamname', 'playername'])

rating_generator = UpdateRatingGenerator(performance_column='performance')

start_rating_search_range = [
    ParameterSearchRange(
        name='team_weight',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='uniform',
        low=20,
        high=100,
    )
]

performance_generator_tuner = PerformancesGeneratorTuner(
    performances_search_range=PerformancesSearchRange(search_ranges=[
        ParameterSearchRange(
            name='damagetochampions',
            type='uniform',
            low=0,
            high=0.45
        ),
        ParameterSearchRange(
            name='deaths',
            type='uniform',
            low=0,
            high=.3,
            lower_is_better=True
        ),
        ParameterSearchRange(
            name='kills',
            type='uniform',
            low=0,
            high=0.3
        ),
        ParameterSearchRange(
            name='result',
            type='uniform',
            low=0.25,
            high=0.85
        ),
    ]),
    n_trials=3
)

rating_generator_tuner = UpdateRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
    start_rating_search_ranges=start_rating_search_range,
    optimize_league_ratings=True,
    team_rating_n_trials=3
)

predictor = GameTeamPredictor(
    game_id_colum="gameid",
    team_id_column="teamname",
)

pipeline = Pipeline(
    rating_generators=rating_generator,
    predictor=predictor,
    column_names=column_names
)

tuner = PipelineTuner(
    performances_generator_tuners=performance_generator_tuner,
    rating_generator_tuners=rating_generator_tuner,
    predictor_tuner=PredictorTuner(n_trials=1, search_ranges=[ParameterSearchRange(name='C', type='categorical', choices=[1.0, 0.5])]),
    fit_best=True,
    pipeline=pipeline,
)
best_match_predictor,df = tuner.tune(df=df, return_df=True, return_cross_validated_predictions=True)



from dataclasses import dataclass
from typing import List, Optional, Union, Any


@dataclass
class ColumnNames:
    team_id: str
    match_id: str
    start_date: str
    player_id: str
    league: Optional[str] = None
    position: Optional[str] = None
    participation_weight: Optional[str] = None
    projected_participation_weight: Optional[str] = None
    update_match_id: Optional[str] = None
    parent_team_id: Optional[str] = None
    team_players_playing_time: Optional[str] = None
    opponent_players_playing_time: Optional[str] = None
    other_values: Optional[list[str]] = None

    def __post_init__(self):
        if self.update_match_id is None:
            self.update_match_id = self.match_id

        if self.parent_team_id is None:
            self.parent_team_id = self.team_id

        if self.update_match_id != self.match_id and self.parent_team_id is None:
            raise ValueError(
                "rating_update_team_id must be passed if rating_update_match_id is passed"
            )

        if self.parent_team_id != self.team_id and self.update_match_id is None:
            raise ValueError(
                "rating_update_match_id must be passed if rating_update_team_id is passed"
            )


@dataclass
class MatchPerformance:
    performance_value: Optional[float]
    participation_weight: Optional[float]
    projected_participation_weight: float
    team_players_playing_time: Optional[dict[str, float]] = None
    opponent_players_playing_time: Optional[dict[str, float]] = None


@dataclass
class PlayerRating:
    id: str
    rating_value: float
    name: Optional[str] = None
    games_played: Union[float, int] = 0
    last_match_day_number: int = None
    confidence_sum: float = 0
    prev_rating_changes: List[float] = None
    most_recent_team_id: Optional[str] = None


@dataclass
class Team:
    id: str
    player_ids: list[str]
    last_match_day_number: int
    name: Optional[str] = None


@dataclass
class TeamRating:
    id: str
    name: str
    rating_value: float
    players: list[PlayerRating]
    games_played: int = 0
    last_match_day_number: int = None


@dataclass
class PreMatchPlayerRating:
    id: str
    rating_value: float
    games_played: int
    league: Optional[str]
    position: Optional[str]
    match_performance: MatchPerformance
    other: Optional[dict[str, Any]] = None


@dataclass
class PreMatchTeamRating:
    id: str
    players: list[PreMatchPlayerRating]
    rating_value: Optional[float]
    projected_rating_value: float
    league: Optional[str]


@dataclass
class PreMatchRating:
    id: str
    teams: list[PreMatchTeamRating]
    day_number: int


@dataclass
class PlayerRatingChange:
    id: str
    day_number: int
    league: Optional[str]
    participation_weight: float
    predicted_performance: float
    performance: float
    pre_match_rating_value: float
    rating_change_value: float


@dataclass
class TeamRatingChange:
    id: str
    players: list[PlayerRatingChange]
    predicted_performance: float
    performance: float
    pre_match_projected_rating_value: float
    rating_change_value: float
    league: Optional[str]


@dataclass
class PostMatchTeamRatingChange:
    id: str
    players: list[PlayerRatingChange]
    rating_value: float
    predicted_performance: float


@dataclass
class PostMatchRatingChange:
    id: str
    teams: list[PostMatchTeamRatingChange]


@dataclass
class MatchRating:
    id: str
    pre_match_rating: PreMatchRating
    post_match_rating: PostMatchRatingChange


@dataclass
class MatchRatings:
    pre_match_team_rating_projected_values: list[float]
    pre_match_team_rating_values: list[float]
    pre_match_player_rating_values: list[float]
    pre_match_opponent_rating_values: list[float]
    player_rating_changes: list[float]
    player_leagues: list[str]
    team_opponent_leagues: list[str]
    match_ids: list[str]


@dataclass
class MatchPlayer:
    id: str
    performance: Optional[MatchPerformance]
    league: Optional[str] = None
    position: Optional[str] = None
    team_players_participation_weight: Optional[dict[str, float]] = None
    opponent_players_participation_weight: Optional[dict[str, float]] = None
    others: Optional[dict[str, Any]] = None


@dataclass
class MatchTeam:
    id: str
    players: list[MatchPlayer]
    league: str = None
    update_id: Optional[str] = None

    def __post_init__(self):

        if self.update_id is None:
            self.update_id = self.id


@dataclass
class Match:
    id: str
    teams: List[MatchTeam]
    day_number: int
    update_id: Optional[str] = None
    league: str = None

    def __post_init__(self):
        if self.update_id is None:
            self.update_id = self.id



from dataclasses import dataclass
from enum import Enum


@dataclass
class InputColumnNames:
    hour = "hour"
    performance = "performance"


@dataclass
class RatingEstimatorFeatures:
    PLAYER_RATING = "player_rating"
    PLAYER_RATING_DIFFERENCE_FROM_TEAM_PROJECTED = (
        "player_rating_difference_from_team_projected"
    )
    OPPONENT_RATING_PROJECTED = "opponent_rating_projected"
    TEAM_RATING_PROJECTED = "team_rating_projected"
    PLAYER_RATING_DIFFERENCE_PROJECTED = "player_rating_difference_projected"
    RATING_DIFFERENCE_PROJECTED = "rating_difference_projected"
    RATING_MEAN_PROJECTED = "rating_mean_projected"
    TEAM_LEAGUE = "team_league"
    PLAYER_LEAGUE = "player_league"
    OPPONENT_LEAGUE = "opponent_league"
    MATCH_ID = "match_id"
    TIME_WEIGHTED_RATING = "time_weighted_rating"
    TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO = "time_weighted_rating_likelihood_ratio"
    TIME_WEIGHTED_RATING_EVIDENCE = "time_weighted_rating_evidence"
    RATING_DIFFERENCE_POSITION = "rating_difference_position"


@dataclass
class RatingHistoricalFeatures:
    PERFORMANCE = "performance"
    RATING_DIFFERENCE = "rating_difference"
    PLAYER_RATING_DIFFERENCE = "player_rating_difference"
    TEAM_RATING = "team_rating"
    OPPONENT_RATING = "opponent_rating"
    PLAYER_RATING_DIFFERENCE_FROM_TEAM = "player_rating_difference_from_team"
    PLAYER_RATING_CHANGE = "player_rating_change"
    RATING_MEAN = "rating_mean"
    PLAYER_PREDICTED_PERFORMANCE = "player_predicted_performance"


class PredictedRatingMethod(Enum):
    DEFAULT = "default"
    MEAN_RATING = "mean_rating"



import copy
import inspect
import logging
from dataclasses import dataclass

import pandas as pd
from player_performance_ratings.data_structures import Match, ColumnNames

from player_performance_ratings.cross_validator._base import CrossValidator

from player_performance_ratings import PipelineFactory
from player_performance_ratings.ratings import UpdateRatingGenerator
from player_performance_ratings.ratings.enums import (
    RatingEstimatorFeatures,
    RatingHistoricalFeatures,
)
from player_performance_ratings.ratings.rating_calculators.start_rating_generator import (
    StartRatingGenerator,
)


@dataclass
class LeagueH2H:
    league: str
    opp_league: str
    weight: float
    mean_rating_change: float


class StartLeagueRatingOptimizer:

    def __init__(
        self,
        pipeline_factory: PipelineFactory,
        cross_validator: CrossValidator,
        max_iterations: int = 10,
        learning_step: int = 20,
        weight_div: int = 500,
        indirect_weight: float = 1.5,
        verbose: int = 1,
    ):
        self.pipeline_factory = pipeline_factory
        self.max_iterations = max_iterations
        self.learning_step = learning_step
        self.weight_div = weight_div
        self.cross_validator = cross_validator
        self.indirect_weight = indirect_weight
        self.verbose = verbose
        self._scores = []
        self._league_ratings_iterations = []

    def optimize(
        self,
        df: pd.DataFrame,
        rating_model_idx: int,
        rating_generator: UpdateRatingGenerator,
        matches: list[Match],
        column_names: ColumnNames,
    ) -> dict[str, float]:

        start_rating_generator = (
            rating_generator.match_rating_generator.start_rating_generator
        )
        league_ratings = start_rating_generator.league_ratings.copy()
        start_rating_params = list(
            inspect.signature(
                start_rating_generator.__class__.__init__
            ).parameters.keys()
        )[1:]
        non_league_ratings_params = {
            attr: getattr(start_rating_generator, attr)
            for attr in dir(start_rating_generator)
            if attr in start_rating_params and attr != "league_ratings"
        }

        for iteration in range(self.max_iterations + 1):

            rating_generator_used = copy.deepcopy(rating_generator)

            start_rating_generator = StartRatingGenerator(
                league_ratings=league_ratings, **non_league_ratings_params
            )

            if iteration == self.max_iterations:
                min_idx = self._scores.index(min(self._scores))
                logging.info(
                    f"best start rating params {self._league_ratings_iterations[min_idx]}"
                )
                return self._league_ratings_iterations[min_idx]

            rating_generator_used.match_rating_generator.start_rating_generator = (
                start_rating_generator
            )

            rating_generators = copy.deepcopy(self.pipeline_factory.rating_generators)
            rating_generators[rating_model_idx] = rating_generator_used

            rating_values = rating_generators[rating_model_idx].generate_historical(
                df=df, matches=matches, column_names=column_names
            )
            for rating_column, rating_value in rating_values.items():
                df = df.assign(**{rating_column: rating_value})

            ratings_df = rating_generators[rating_model_idx].ratings_df
            df_incl_ratings = df.merge(
                ratings_df[
                    [
                        column_names.match_id,
                        column_names.team_id,
                        column_names.player_id,
                        RatingHistoricalFeatures.PLAYER_RATING_CHANGE,
                        RatingEstimatorFeatures.PLAYER_LEAGUE,
                        RatingEstimatorFeatures.OPPONENT_LEAGUE,
                    ]
                ],
                on=[
                    column_names.match_id,
                    column_names.player_id,
                    column_names.team_id,
                ],
                how="inner",
            )

            rating_generators = copy.deepcopy(self.pipeline_factory.rating_generators)
            rating_generators[
                rating_model_idx
            ].match_rating_generator.start_rating_generator = start_rating_generator
            pipeline = self.pipeline_factory.create(rating_generators=rating_generators)
            score = pipeline.cross_validate_score(
                df=df_incl_ratings,
                cross_validator=self.cross_validator,
                create_performance=False,
                create_rating_features=False,
            )

            if not league_ratings:
                league_ratings = pipeline.rating_generators[
                    rating_model_idx
                ].match_rating_generator.start_rating_generator.league_ratings

            self._league_ratings_iterations.append(copy.deepcopy(league_ratings))
            self._scores.append(score)
            if self.verbose:
                logging.info(
                    f"iteration {iteration} finished. Score: {score}. best startings {league_ratings}"
                )

            league_rating_changes = (
                df_incl_ratings.groupby(
                    [
                        RatingEstimatorFeatures.PLAYER_LEAGUE,
                        RatingEstimatorFeatures.OPPONENT_LEAGUE,
                    ]
                )
                .agg(
                    {
                        RatingHistoricalFeatures.PLAYER_RATING_CHANGE: "mean",
                        column_names.player_id: "count",
                    }
                )
                .reset_index()
                .rename(
                    columns={
                        RatingHistoricalFeatures.PLAYER_RATING_CHANGE: "mean_rating_change",
                        column_names.player_id: "count",
                    }
                )
            )
            leagues = (
                league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE]
                .unique()
                .tolist()
            )

            league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}
            league_to_played_against_leagues = {}

            for league in leagues:
                league_to_played_against_leagues[league] = (
                    league_rating_changes[
                        league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE]
                        == league
                    ][RatingEstimatorFeatures.OPPONENT_LEAGUE]
                    .unique()
                    .tolist()
                )

                league_opp_league_h2hs[league] = {}
                for opp_league in leagues:
                    if league == opp_league:
                        continue

                    rows = league_rating_changes[
                        (
                            league_rating_changes[RatingEstimatorFeatures.PLAYER_LEAGUE]
                            == league
                        )
                        & (
                            league_rating_changes[
                                RatingEstimatorFeatures.OPPONENT_LEAGUE
                            ]
                            == opp_league
                        )
                    ]
                    if len(rows) == 0:
                        weight = 0
                        mean_rating_change = 0
                    else:
                        count = rows["count"].iloc[0]
                        weight = min(1, count / self.weight_div)
                        mean_rating_change = rows["mean_rating_change"].iloc[0]

                    league_h2h = LeagueH2H(
                        weight=weight,
                        mean_rating_change=mean_rating_change,
                        league=league,
                        opp_league=opp_league,
                    )
                    league_opp_league_h2hs[league][opp_league] = league_h2h

            final_league_opp_league_h2hs: dict[str, dict[str, LeagueH2H]] = {}

            league_sum_final_weights = {}
            for league, opp_league_h2h in league_opp_league_h2hs.items():

                league_sum_final_weights[league] = 0
                final_league_opp_league_h2hs[league] = {}
                for opp_league, h2h in opp_league_h2h.items():

                    if h2h.weight < 1:
                        league_against = league_to_played_against_leagues[league]
                        opp_played_against = league_to_played_against_leagues[
                            opp_league
                        ]
                        shared_leagues = [
                            l
                            for l in league_against
                            if l in opp_played_against and l not in (league, opp_league)
                        ]

                        relative_shared_weighted_mean = 0
                        shared_weight = 0
                        for shared_league in shared_leagues:
                            opp_vs_shared_h2h = league_opp_league_h2hs[opp_league][
                                shared_league
                            ]
                            league_vs_shared_h2h = league_opp_league_h2hs[league][
                                shared_league
                            ]

                            sum_rating_change = (
                                league_vs_shared_h2h.mean_rating_change
                                - opp_vs_shared_h2h.mean_rating_change
                            )

                            relative_shared_weighted_mean += (
                                min(
                                    opp_vs_shared_h2h.weight,
                                    league_vs_shared_h2h.weight,
                                )
                                * sum_rating_change
                            )
                            shared_weight += min(
                                opp_vs_shared_h2h.weight, league_vs_shared_h2h.weight
                            )

                    weighted_mean = (
                        h2h.mean_rating_change * h2h.weight
                        + min(1, shared_weight * self.indirect_weight)
                        * relative_shared_weighted_mean
                    )
                    final_weight = min(1, h2h.weight + shared_weight)
                    final_league_h2h = LeagueH2H(
                        weight=final_weight,
                        mean_rating_change=weighted_mean,
                        league=league,
                        opp_league=opp_league,
                    )
                    final_league_opp_league_h2hs[league][opp_league] = final_league_h2h
                    league_sum_final_weights[league] += final_weight

            for league, final_opp_league_h2h in final_league_opp_league_h2hs.items():

                if league not in league_ratings:
                    continue
                start_rating_rating_change = 0

                for opp_league, final_h2h in final_opp_league_h2h.items():
                    if league_sum_final_weights[league] > 0:
                        start_rating_rating_change += (
                            final_h2h.weight
                            / league_sum_final_weights[league]
                            * final_h2h.mean_rating_change
                            * self.learning_step
                        )

                league_ratings[league] += start_rating_rating_change

        return league_ratings



from .enums import (
    RatingEstimatorFeatures,
    RatingHistoricalFeatures,
    PredictedRatingMethod,
    InputColumnNames,
)
from .league_identifier import LeagueIdentifier
from .match_generator import convert_df_to_matches
from .update_rating_generator import UpdateRatingGenerator
from .rating_calculators.match_rating_generator import MatchRatingGenerator
from .rating_calculators.start_rating_generator import StartRatingGenerator
from .rating_calculators.performance_predictor import (
    RatingDifferencePerformancePredictor,
    RatingMeanPerformancePredictor,
    RatingNonOpponentPerformancePredictor,
)



from abc import ABC, abstractmethod
from typing import Optional, Match

import pandas as pd


class BaseTuner(ABC):

    @abstractmethod
    def tune(
        self, df: pd.DataFrame, matches: Optional[list[Match]] = None
    ) -> dict[str, float]:
        pass



import copy
import inspect
from typing import Optional

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial
from player_performance_ratings.transformers.base_transformer import BaseTransformer

from player_performance_ratings.cross_validator.cross_validator import CrossValidator
from player_performance_ratings import PipelineFactory, ColumnNames

from player_performance_ratings.predictor import BasePredictor


from player_performance_ratings.tuner.utils import (
    ParameterSearchRange,
    add_params_from_search_range,
    get_default_lgbm_regressor_search_range,
)


class PredictorTuner:

    def __init__(
        self,
        search_ranges: Optional[list[ParameterSearchRange]] = None,
        default_params: Optional[dict] = None,
        n_trials: int = 30,
    ):
        self.search_ranges = search_ranges
        self.default_params = default_params or {}
        self.n_trials = n_trials

    def tune(
        self,
        df: pd.DataFrame,
        pipeline_factory: PipelineFactory,
        cross_validator: CrossValidator,
    ) -> BasePredictor:

        deepest_estimator = pipeline_factory.predictor.estimator

        estimator_subclass_level = 0
        while hasattr(deepest_estimator, "estimator"):
            deepest_estimator = deepest_estimator.estimator
            estimator_subclass_level += 1

        if self.search_ranges is None and deepest_estimator.__class__.__name__ in (
            "LGBMRegressor",
            "LGBMClassifier",
        ):
            search_ranges = get_default_lgbm_regressor_search_range()
        elif self.search_ranges is None:
            raise ValueError(
                "search_ranges can't be None if estimator is not LGBMRegressor or LGBMClassifier"
            )
        else:
            search_ranges = self.search_ranges

        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:

            predictor = pipeline_factory.predictor

            if estimator_subclass_level == 0:
                param_names = list(
                    inspect.signature(
                        predictor.estimator.__class__.__init__
                    ).parameters.keys()
                )[1:]
                params = {
                    attr: getattr(predictor.estimator, attr)
                    for attr in param_names
                    if attr != "kwargs"
                }
                if "_other_params" in predictor.estimator.__dict__:
                    params.update(predictor.estimator._other_params)
            elif estimator_subclass_level == 1:
                param_names = list(
                    inspect.signature(
                        predictor.estimator.estimator.__class__.__init__
                    ).parameters.keys()
                )[1:]
                params = {
                    attr: getattr(predictor.estimator.estimator, attr)
                    for attr in param_names
                    if attr != "kwargs"
                }
                if "_other_params" in predictor.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator._other_params)
            elif estimator_subclass_level == 2:
                param_names = list(
                    inspect.signature(
                        predictor.estimator.estimator.estimator.__class__.__init__
                    ).parameters.keys()
                )[1:]
                params = {
                    attr: getattr(predictor.estimator.estimator.estimator, attr)
                    for attr in param_names
                    if attr != "kwargs"
                }
                if "_other_params" in predictor.estimator.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator.estimator._other_params)

            else:
                raise ValueError(
                    f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}"
                )

            params = add_params_from_search_range(
                params=params, trial=trial, parameter_search_range=search_ranges
            )
            for param, value in self.default_params.items():
                params[param] = value

            predictor = copy.deepcopy(pipeline_factory.predictor)
            for param in params:
                if estimator_subclass_level == 1:
                    setattr(predictor.estimator.estimator, param, params[param])
                elif estimator_subclass_level == 2:
                    setattr(
                        predictor.estimator.estimator.estimator, param, params[param]
                    )
                elif estimator_subclass_level > 2:
                    raise ValueError(
                        f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}"
                    )
                else:
                    setattr(predictor.estimator, param, params[param])

            pipeline = pipeline_factory.create(predictor=predictor)
            create_rating_features = False
            for rating_generator in pipeline.rating_generators:
                create_rating_features = any(
                    feature not in df.columns
                    for feature in rating_generator.estimator_features_return
                )
                if create_rating_features:
                    break
            return pipeline.cross_validate_score(
                df=df,
                create_performance=False,
                create_rating_features=create_rating_features,
                cross_validator=cross_validator,
            )

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(
            direction=direction, study_name=study_name, sampler=sampler
        )
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df),
            n_trials=self.n_trials,
            callbacks=callbacks,
        )
        best_estimator_params = study.best_params
        other_predictor_params = list(
            inspect.signature(
                pipeline_factory.predictor.__class__.__init__
            ).parameters.keys()
        )[1:]

        if estimator_subclass_level > 0:
            if estimator_subclass_level == 1:
                best_estimator_params.update(
                    pipeline_factory.predictor.estimator.estimator._other_params
                )
            elif estimator_subclass_level == 2:
                best_estimator_params.update(
                    pipeline_factory.predictor.estimator.estimator.estimator._other_params
                )

        else:
            if "_other_params" in pipeline_factory.predictor.estimator.__dict__:
                best_estimator_params.update(
                    pipeline_factory.predictor.estimator._other_params
                )

        other_predictor_params = {
            attr: getattr(pipeline_factory.predictor, attr)
            for attr in other_predictor_params
            if attr not in ("estimator")
        }

        predictor_class = pipeline_factory.predictor.__class__
        if estimator_subclass_level == 1:

            potential_parent_names = list(
                inspect.signature(
                    pipeline_factory.predictor.estimator.__class__.__init__
                ).parameters.keys()
            )[1:]
            other_parent_params = {
                attr: getattr(pipeline_factory.predictor.estimator, attr)
                for attr in potential_parent_names
                if attr != "estimator"
            }

            estimator_class = pipeline_factory.predictor.estimator.estimator.__class__
            parent_estimator_class = pipeline_factory.predictor.estimator.__class__
            parent_estimator = parent_estimator_class(
                estimator=estimator_class(**best_estimator_params),
                **other_parent_params,
            )
            return predictor_class(estimator=parent_estimator, **other_predictor_params)
        elif estimator_subclass_level == 2:
            potential_parent_names = list(
                inspect.signature(
                    pipeline_factory.predictor.estimator.estimator.__class__.__init__
                ).parameters.keys()
            )[1:]
            other_parent_params = {
                attr: getattr(pipeline_factory.predictor.estimator.estimator, attr)
                for attr in potential_parent_names
                if attr != "estimator"
            }
            estimator_class = (
                pipeline_factory.predictor.estimator.estimator.estimator.__class__
            )
            parent_estimator_class = (
                pipeline_factory.predictor.estimator.estimator.__class__
            )
            parent_estimator = parent_estimator_class(
                estimator=estimator_class(**best_estimator_params),
                **other_parent_params,
            )

            parent_parent_estimator_class = (
                pipeline_factory.predictor.estimator.__class__
            )
            parent_parent_estimator = parent_parent_estimator_class(
                estimator=parent_estimator
            )
            return predictor_class(
                estimator=parent_parent_estimator, **other_predictor_params
            )

        elif estimator_subclass_level == 0:

            estimator_class = pipeline_factory.predictor.estimator.__class__
            return predictor_class(
                estimator=estimator_class(**best_estimator_params),
                **other_predictor_params,
            )

        else:
            raise ValueError(
                f"estimator_subclass_level can't be higher than 2, got {estimator_subclass_level}"
            )



from .performance_predictor import (
    RatingDifferencePerformancePredictor,
    RatingMeanPerformancePredictor,
)
from .start_rating_generator import StartRatingGenerator
from .match_rating_generator import MatchRatingGenerator



from abc import abstractmethod, ABC
from typing import Optional

import numpy as np
import pandas as pd
import polars as pl
from player_performance_ratings import ColumnNames


class BasePerformancesTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


class BaseTransformer(ABC):

    def __init__(
        self,
        features: list[str],
        features_out: list[str],
        are_estimator_features: bool = True,
    ):
        self._features_out = features_out
        self.features = features
        self._are_estimator_features = are_estimator_features
        self.column_names = None
        self._estimator_features_out = (
            self._features_out if self._are_estimator_features else []
        )

    @abstractmethod
    def fit_transform(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    @property
    def estimator_features_out(self) -> list[str]:
        return self._estimator_features_out

    def reset(self) -> "BaseTransformer":
        return self


class BaseLagGenerator:

    def __init__(
        self,
        granularity: list[str],
        features: list[str],
        add_opponent: bool,
        iterations: list[int],
        prefix: str,
        are_estimator_features: bool = True,
    ):

        self.features = features
        self.iterations = iterations
        self._features_out = []
        self._are_estimator_features = are_estimator_features
        self.granularity = granularity
        if isinstance(self.granularity, str):
            self.granularity = [self.granularity]
        self._entity_features = []
        self.add_opponent = add_opponent
        self.prefix = prefix
        self._df = None
        self._entity_features = []
        self.column_names: Optional[ColumnNames] = None

        for feature_name in self.features:
            for lag in iterations:
                self._features_out.append(f"{prefix}{lag}_{feature_name}")
                self._entity_features.append(f"{prefix}{lag}_{feature_name}")
                if self.add_opponent:
                    self._features_out.append(f"{prefix}{lag}_{feature_name}_opponent")

    @abstractmethod
    def generate_historical(
        self, df: pd.DataFrame, column_names: ColumnNames
    ) -> pd.DataFrame:
        pass

    @abstractmethod
    def generate_future(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def estimator_features_out(self) -> list[str]:
        if self._are_estimator_features:
            return self.features_out
        return []

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def _concat_df(
        self, df: pd.DataFrame, additional_cols_to_use: Optional[list[str]] = None
    ) -> pd.DataFrame:
        df = self._string_convert(df=df)
        for feature in self.features:
            if feature in df.columns:
                df = df.assign(**{feature: lambda x: x[feature].astype("float")})

        df = df[[c for c in df.columns if c not in self.features_out]]

        cols = [
            f
            for f in list(
                set(
                    [
                        *self.features,
                        *self.granularity,
                        self.column_names.match_id,
                        self.column_names.team_id,
                        "is_future",
                        self.column_names.player_id,
                        self.column_names.parent_team_id,
                        self.column_names.update_match_id,
                        self.column_names.start_date,
                    ]
                )
            )
            if f in df.columns
        ]

        if self.column_names.participation_weight in df.columns:
            cols += [self.column_names.participation_weight]
        if self.column_names.projected_participation_weight in df.columns:
            cols += [self.column_names.projected_participation_weight]

        if additional_cols_to_use:
            cols += [f for f in additional_cols_to_use if f in df.columns]

        concat_df = pd.concat([self._df, df[cols]], axis=0).reset_index()

        if "index" in concat_df.columns:
            concat_df = concat_df.drop(columns=["index"])
        if concat_df[self.column_names.start_date].dtype in ("str", "object"):
            concat_df[self.column_names.start_date] = pd.to_datetime(
                concat_df[self.column_names.start_date]
            )
        return concat_df.drop_duplicates(
            subset=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            keep="last",
        )

    def _store_df(
        self, df: pd.DataFrame, additional_cols_to_use: Optional[list[str]] = None
    ):
        df = self._string_convert(df)

        cols = list(
            set(
                [
                    *self.features,
                    *self.granularity,
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                    self.column_names.parent_team_id,
                    self.column_names.update_match_id,
                    self.column_names.start_date,
                    "is_future",
                ]
            )
        )
        if self.column_names.participation_weight in df.columns:
            cols += [self.column_names.participation_weight]
        if self.column_names.projected_participation_weight in df.columns:
            cols += [self.column_names.projected_participation_weight]

        if additional_cols_to_use:
            cols += additional_cols_to_use

        if self._df is None:
            self._df = df[cols]
        else:
            self._df = pd.concat([self._df, df[cols]], axis=0)

        self._df = self._df.drop_duplicates(
            subset=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            keep="last",
        )

    def _string_convert(self, df: pd.DataFrame) -> pd.DataFrame:
        for column in [
            self.column_names.match_id,
            self.column_names.parent_team_id,
            self.column_names.player_id,
            self.column_names.update_match_id,
        ]:
            df = df.assign(**{column: lambda x: x[column].astype("str")})
        return df

    def _create_transformed_df(
        self, df: pd.DataFrame, concat_df: pd.DataFrame
    ) -> pd.DataFrame:

        cn = self.column_names

        if self.add_opponent:
            concat_df = self._add_opponent_features(df=concat_df)

        ori_cols = [c for c in df.columns if c not in concat_df.columns] + [
            cn.match_id,
            cn.player_id,
            cn.team_id,
        ]
        ori_index_values = df.index.tolist()

        df = self._string_convert(df)

        transformed_df = concat_df.merge(
            df[ori_cols], on=[cn.match_id, cn.player_id, cn.team_id], how="inner"
        )

        transformed_df.index = ori_index_values
        transformed_df = transformed_df.sort_values(
            by=[cn.start_date, cn.match_id, cn.team_id, cn.player_id]
        )

        return transformed_df[list(set(df.columns.tolist() + self.features_out))]

    def _add_opponent_features(self, df: pd.DataFrame) -> pd.DataFrame:
        team_features = (
            df.groupby([self.column_names.team_id, self.column_names.match_id])[
                self._entity_features
            ]
            .mean()
            .reset_index()
        )
        df_opponent_feature = team_features.rename(
            columns={
                **{self.column_names.team_id: "__opponent_team_id"},
                **{f: f"{f}_opponent" for f in self._entity_features},
            }
        )
        new_df = df.merge(
            df_opponent_feature,
            on=[self.column_names.match_id],
            suffixes=("", "_team_sum"),
        )
        new_df = new_df[
            new_df[self.column_names.team_id] != new_df["__opponent_team_id"]
        ].drop(columns=["__opponent_team_id"])

        new_feats = [f"{f}_opponent" for f in self._entity_features]
        return df.merge(
            new_df[
                [
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                    *new_feats,
                ]
            ],
            on=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            how="left",
        )

    def _generate_future_feats(
        self,
        transformed_df: pd.DataFrame,
        ori_df: pd.DataFrame,
        known_future_features: Optional[list[str]] = None,
    ) -> pd.DataFrame:
        known_future_features = known_future_features or []
        ori_cols = ori_df.columns.tolist()
        ori_index_values = ori_df.index.tolist()
        cn = self.column_names

        transformed_df[self._entity_features] = transformed_df[
            self._entity_features
        ].fillna(-999.21345)
        first_grp = (
            transformed_df.groupby(self.granularity)[self._entity_features]
            .first()
            .reset_index()
        )
        transformed_df = transformed_df[
            [c for c in transformed_df.columns if c not in self._entity_features]
        ].merge(first_grp, on=self.granularity, how="left")
        for f in self._entity_features:
            transformed_df[f].replace(-999.21345, np.nan, inplace=True)
        if not transformed_df[self._entity_features].isnull().all().all():
            transformed_df[self._entity_features] = transformed_df.groupby(
                self.granularity
            )[self._entity_features].transform(lambda x: x.fillna(method="ffill"))

        team_features = (
            transformed_df.groupby(
                [self.column_names.team_id, self.column_names.match_id]
            )[self._entity_features]
            .mean()
            .reset_index()
        )
        df_opponent_feature = team_features.rename(
            columns={
                **{self.column_names.team_id: "__opponent_team_id"},
                **{f: f"{f}_opponent" for f in self._entity_features},
            }
        )
        opponent_feat_names = [f"{f}_opponent" for f in self._entity_features]
        new_df = transformed_df.merge(
            df_opponent_feature,
            on=[self.column_names.match_id],
            suffixes=("", "_team_sum"),
        )
        new_df = new_df[
            new_df[self.column_names.team_id] != new_df["__opponent_team_id"]
        ]
        new_df[opponent_feat_names] = new_df[opponent_feat_names].fillna(-999.21345)
        first_grp = (
            new_df.groupby("__opponent_team_id")[opponent_feat_names]
            .first()
            .reset_index()
        )
        new_df = new_df[
            [c for c in new_df.columns if c not in opponent_feat_names]
        ].merge(first_grp, on="__opponent_team_id", how="left")
        for f in opponent_feat_names:
            new_df[f].replace(-999.21345, np.nan, inplace=True)

        new_df = new_df.sort_values(
            by=[cn.start_date, cn.match_id, "__opponent_team_id"]
        )
        new_df.groupby("__opponent_team_id")[opponent_feat_names].fillna(
            method="ffill", inplace=True
        )

        transformed_df = transformed_df.merge(
            new_df[
                [
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                    *opponent_feat_names,
                ]
            ],
            on=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            how="left",
        )

        for col in [cn.match_id, cn.player_id, cn.team_id]:
            ori_df = ori_df.assign(**{col: lambda x: x[col].astype("str")})

        ori_feats_to_use = [f for f in ori_cols if f not in self.features_out]
        transformed_feats_to_use = [
            cn.match_id,
            cn.team_id,
            cn.player_id,
            *[f for f in self.features_out if f not in known_future_features],
        ]

        transformed_df = ori_df[ori_feats_to_use].merge(
            transformed_df[transformed_feats_to_use],
            on=[cn.match_id, cn.team_id, cn.player_id],
        )

        transformed_df = transformed_df.sort_values(
            by=[cn.start_date, cn.match_id, cn.team_id, cn.player_id]
        )
        transformed_df.index = ori_index_values
        return transformed_df[
            list(
                set(
                    ori_cols
                    + [f for f in self.features_out if f not in known_future_features]
                )
            )
        ]

    def reset(self) -> "BaseLagGenerator":
        self._df = None
        return self


class BaseLagGeneratorPolars:

    def __init__(
        self,
        granularity: list[str],
        features: list[str],
        add_opponent: bool,
        iterations: list[int],
        prefix: str,
        are_estimator_features: bool = True,
    ):

        self.features = features
        self.iterations = iterations
        self._features_out = []
        self._are_estimator_features = are_estimator_features
        self.granularity = granularity
        if isinstance(self.granularity, str):
            self.granularity = [self.granularity]
        self._entity_features = []
        self.add_opponent = add_opponent
        self.prefix = prefix
        self._df = None
        self._entity_features = []
        self.column_names: Optional[ColumnNames] = None

        for feature_name in self.features:
            for lag in iterations:
                self._features_out.append(f"{prefix}{lag}_{feature_name}")
                self._entity_features.append(f"{prefix}{lag}_{feature_name}")
                if self.add_opponent:
                    self._features_out.append(f"{prefix}{lag}_{feature_name}_opponent")

    @abstractmethod
    def generate_historical(
        self, df: pl.DataFrame, column_names: ColumnNames
    ) -> pl.DataFrame:
        pass

    @abstractmethod
    def generate_future(self, df: pl.DataFrame) -> pl.DataFrame:
        pass

    @property
    def estimator_features_out(self) -> list[str]:
        if self._are_estimator_features:
            return self.features_out
        return []

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def _concat_df(self, df: pl.DataFrame) -> pl.DataFrame:
        df = self._string_convert(df=df)
        df = df.with_columns(
            [
                pl.col(feature).cast(pl.Float64).alias(feature)
                for feature in self.features
                if feature in df.columns
            ]
        )

        df = df[[c for c in df.columns if c not in self.features_out]]
        cols = [c for c in self._df.columns if c in df.columns]

        concat_df = pl.concat([self._df, df.select(cols)], how="diagonal_relaxed")

        if concat_df[self.column_names.start_date].dtype in ("str", "object"):
            concat_df[self.column_names.start_date] = pd.to_datetime(
                concat_df[self.column_names.start_date]
            )
        return concat_df.unique(
            subset=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
            # maintain_order=True
        )

    def _store_df(
        self, df: pl.DataFrame, additional_cols_to_use: Optional[list[str]] = None
    ):
        df = df.with_columns(
            [
                pl.col(feature).cast(pl.Float64).alias(feature)
                for feature in self.features
                if feature in df.columns
            ]
        )

        df = self._string_convert(df)

        cols = list(
            {
                *self.features,
                *self.granularity,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
                "is_future",
                self.column_names.parent_team_id,
                self.column_names.update_match_id,
                self.column_names.start_date,
            }
        )
        if self.column_names.participation_weight in df.columns:
            cols += [self.column_names.participation_weight]
        if self.column_names.projected_participation_weight in df.columns:
            cols += [self.column_names.projected_participation_weight]

        if additional_cols_to_use:
            cols += additional_cols_to_use

        if self._df is None:
            self._df = df.select(cols)
        else:
            self._df = pl.concat([self._df, df.select(cols)])

        self._df = self._df.sort(
            [
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            #   descending=True
        ).unique(
            subset=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            #  maintain_order=True
        )

    def _string_convert(self, df: pl.DataFrame) -> pl.DataFrame:
        for column in [
            self.column_names.match_id,
            self.column_names.parent_team_id,
            self.column_names.player_id,
            self.column_names.update_match_id,
        ]:
            df = df.with_columns(df[column].cast(pl.Utf8))
        return df

    def _create_transformed_df(
        self, df: pl.DataFrame, concat_df: pl.DataFrame
    ) -> pl.DataFrame:

        cn = self.column_names

        if self.add_opponent:
            concat_df = self._add_opponent_features(df=concat_df)

        ori_cols = [c for c in df.columns if c not in concat_df.columns] + [
            cn.match_id,
            cn.player_id,
            cn.team_id,
        ]

        df = self._string_convert(df)

        transformed_df = concat_df.join(
            df.select(ori_cols), on=[cn.match_id, cn.player_id, cn.team_id], how="inner"
        )
        return transformed_df.select(list(set(df.columns + self.features_out)))

    def _add_opponent_features(self, df: pl.DataFrame) -> pl.DataFrame:
        team_features = df.groupby(
            [self.column_names.team_id, self.column_names.match_id]
        ).agg(**{col: pl.mean(col) for col in self._entity_features})

        df_opponent_feature = team_features.with_columns(
            [
                pl.col(self.column_names.team_id).alias("__opponent_team_id"),
                *[pl.col(f).alias(f"{f}_opponent") for f in self._entity_features],
            ]
        )

        new_df = df.join(
            df_opponent_feature, on=self.column_names.match_id, suffix="_team_sum"
        )

        new_df = new_df.filter(
            pl.col(self.column_names.team_id) != pl.col("__opponent_team_id")
        ).drop(["__opponent_team_id"])

        new_feats = [f"{f}_opponent" for f in self._entity_features]
        return df.join(
            new_df[
                [
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                    *new_feats,
                ]
            ],
            on=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            how="left",
        )

    def _generate_future_feats(
        self,
        transformed_df: pl.DataFrame,
        ori_df: pl.DataFrame,
        known_future_features: Optional[list[str]] = None,
    ) -> pl.DataFrame:
        known_future_features = known_future_features or []
        ori_cols = ori_df.columns
        cn = self.column_names

        transformed_df[self._entity_features] = transformed_df[
            self._entity_features
        ].fill_nan(-999.21345)
        first_grp = transformed_df.group_by(self.granularity).agg(
            [pl.col(f).first().alias(f) for f in self._entity_features]
        )
        transformed_df = transformed_df[
            [c for c in transformed_df.columns if c not in self._entity_features]
        ].join(first_grp, on=self.granularity, how="left")

        transformed_df = transformed_df.sort([cn.start_date, cn.match_id, cn.team_id])
        for f in self._entity_features:
            transformed_df = transformed_df.with_columns(
                pl.when(pl.col(f) == -999.21345)
                .then(np.nan)
                .otherwise(pl.col(f))
                .alias(f)
            )
            if transformed_df[f].is_null().sum() == len(transformed_df):
                transformed_df = transformed_df.with_columns(
                    pl.col(f).forward_fill().over(self.granularity).alias(f)
                )

        team_features = transformed_df.group_by(
            [self.column_names.team_id, self.column_names.match_id]
        ).agg([pl.col(f).mean().alias(f) for f in self._entity_features])

        rename_mapping = {
            self.column_names.team_id: "__opponent_team_id",
            **{f: f"{f}_opponent" for f in self._entity_features},
        }

        df_opponent_feature = team_features.select(
            [
                pl.col(name).alias(rename_mapping.get(name, name))
                for name in team_features.columns
            ]
        )
        opponent_feat_names = [f"{f}_opponent" for f in self._entity_features]
        new_df = transformed_df.join(
            df_opponent_feature, on=[self.column_names.match_id], suffix="_team_sum"
        )
        new_df = new_df.filter(
            new_df[self.column_names.team_id] != new_df["__opponent_team_id"]
        )
        new_df = new_df.with_columns(
            [
                pl.col(column).fill_nan(-999.21345).alias(column)
                for column in opponent_feat_names
            ]
        )
        first_grp = new_df.group_by("__opponent_team_id").agg(
            [pl.col(f).first().alias(f) for f in opponent_feat_names]
        )
        new_df = new_df[
            [c for c in new_df.columns if c not in opponent_feat_names]
        ].join(first_grp, on="__opponent_team_id", how="left")
        new_df = new_df.with_columns(
            [
                pl.when(pl.col(f) == -999.21345)
                .then(np.nan)
                .otherwise(pl.col(f))
                .alias(f)
                for f in opponent_feat_names
            ]
        )

        new_df = new_df.sort([cn.start_date, cn.match_id, "__opponent_team_id"])
        for f in opponent_feat_names:
            new_df = new_df.with_columns(
                pl.col(f).forward_fill().over("__opponent_team_id").alias(f)
            )

        transformed_df = transformed_df.join(
            new_df.select(
                [
                    self.column_names.match_id,
                    self.column_names.team_id,
                    self.column_names.player_id,
                    *opponent_feat_names,
                ]
            ),
            on=[
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ],
            how="left",
        )

        ori_df = ori_df.with_columns(
            [
                pl.col(col).cast(pl.Utf8).alias(col)
                for col in [cn.match_id, cn.player_id, cn.team_id]
            ]
        )

        ori_feats_to_use = [f for f in ori_cols if f not in self.features_out]
        transformed_feats_to_use = [
            cn.match_id,
            cn.team_id,
            cn.player_id,
            *[f for f in self.features_out if f not in known_future_features],
        ]

        transformed_df = ori_df[ori_feats_to_use].join(
            transformed_df[transformed_feats_to_use],
            on=[cn.match_id, cn.team_id, cn.player_id],
        )

        return transformed_df.select(
            list(
                set(
                    ori_cols
                    + [f for f in self.features_out if f not in known_future_features]
                )
            )
        )

    def reset(self) -> "BaseLagGeneratorPolars":
        self._df = None
        return self



import copy
import logging
import warnings
import pandas as pd
from lightgbm import LGBMClassifier, LGBMRegressor
from pandas.errors import SettingWithCopyWarning
from sklearn import clone

from player_performance_ratings.predictor.sklearn_estimator import OrdinalClassifier
from player_performance_ratings.predictor_transformer import PredictorTransformer
from player_performance_ratings.scorer.score import Filter, apply_filters, Operator

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from typing import Optional

from sklearn.linear_model import LogisticRegression

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor._base import BasePredictor


class GameTeamPredictor(BasePredictor):

    def __init__(
        self,
        game_id_colum: str,
        team_id_column: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
        filters: Optional[list[Filter]] = None,
    ):
        """
        Wrapper for sklearn models that predicts game results.

        The GameTeam Predictor is intended to transform predictions from a lower granularity into a GameTeam level.
        So if input data is at game-player, data is converted to game_team before being trained
        Similar concept if it is at a granularity below game level.


        :param game_id_colum:
        :param team_id_column:
        :param features:
        :param target:
        :param estimator:
        :param multiclassifier:
        :param pred_column:
        """

        self.game_id_colum = game_id_colum
        self.team_id_column = team_id_column
        self._target = target
        self._estimator_features = []

        self.multiclassifier = multiclassifier
        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator or LogisticRegression(),
            pre_transformers=pre_transformers,
            estimator_features=estimator_features,
            filters=filters,
        )

    def train(
        self, df: pd.DataFrame, estimator_features: list[Optional[str]] = None
    ) -> None:

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )

        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()
        df = apply_filters(df=df, filters=self.filters)
        df = self.fit_transform_pre_transformers(df=df)
        if len(df[self._target].unique()) > 2 and hasattr(
            self.estimator, "predict_proba"
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype("int")})
            except Exception:
                pass

        if self._target not in df.columns:
            raise ValueError(f"target {self._target} not in df")

        grouped = self._create_grouped(df)
        self.estimator.fit(grouped[self._estimator_features], grouped[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """

        if hasattr(self.estimator, "predict_proba"):
            try:
                df[self._target] = df[self._target].astype("int")
            except Exception:
                pass
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")
        df = self.transform_pre_transformers(df=df)
        grouped = self._create_grouped(df)

        if self.multiclassifier:
            grouped[self._pred_column] = self.estimator.predict_proba(
                grouped[self._estimator_features]
            ).tolist()
            grouped["classes"] = [
                list(self.estimator.classes_) for _ in range(len(grouped))
            ]
        elif not hasattr(self.estimator, "predict_proba"):
            grouped[self._pred_column] = self.estimator.predict(
                grouped[self._estimator_features]
            )
        else:
            grouped[self._pred_column] = self.estimator.predict_proba(
                grouped[self._estimator_features]
            )[:, 1]

        if self.pred_column in df.columns:
            df = df.drop(columns=[self.pred_column])

        if "classes" in grouped.columns:
            df = df.merge(
                grouped[
                    [self.game_id_colum, self.team_id_column]
                    + [self._pred_column, "classes"]
                ],
                on=[self.game_id_colum, self.team_id_column],
            )

        else:
            df = df.merge(
                grouped[
                    [self.game_id_colum, self.team_id_column] + [self._pred_column]
                ],
                on=[self.game_id_colum, self.team_id_column],
            )

        return df

    def _create_grouped(self, df: pd.DataFrame) -> pd.DataFrame:

        numeric_features = [
            feature
            for feature in self._estimator_features
            if df[feature].dtype in ["int", "float"]
        ]
        cat_feats = [
            feature
            for feature in self._estimator_features
            if feature not in numeric_features
        ]

        if self._target in df.columns:
            if df[self._target].dtype == "object":
                df.loc[:, self._target] = df[self._target].astype("int")

            grouped = (
                df.groupby([self.game_id_colum, self.team_id_column])
                .agg(
                    {
                        **{feature: "mean" for feature in numeric_features},
                        self._target: "mean",
                    }
                )
                .reset_index()
            )

        else:
            grouped = (
                df.groupby([self.game_id_colum, self.team_id_column])
                .agg({**{feature: "mean" for feature in numeric_features}})
                .reset_index()
            )

        if self._target in df.columns and hasattr(
            self._deepest_estimator, "predict_proba"
        ):
            grouped[self._target] = grouped[self._target].astype("int")

        grouped = grouped.merge(
            df[[self.game_id_colum, self.team_id_column, *cat_feats]].drop_duplicates(
                subset=[self.game_id_colum, self.team_id_column]
            ),
            on=[self.game_id_colum, self.team_id_column],
            how="inner",
        )

        return grouped


class Predictor(BasePredictor):

    def __init__(
        self,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.multiclassifier = multiclassifier
        self.column_names = column_names

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100, learning_rate=0.1)"
            )

        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(
        self, df: pd.DataFrame, estimator_features: Optional[list[str]] = None
    ) -> None:

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )
        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()

        filtered_df = apply_filters(df=df, filters=self.filters)
        if hasattr(self.estimator, "predict_proba"):
            try:
                filtered_df[self._target] = filtered_df[self._target].astype("int")
            except Exception:
                pass

        filtered_df = self.fit_transform_pre_transformers(df=filtered_df)

        if (
            not self.multiclassifier
            and len(filtered_df[self._target].unique()) > 2
            and hasattr(self._deepest_estimator, "predict_proba")
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)
            if len(filtered_df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets"
                )

        if hasattr(self._deepest_estimator, "predict_proba"):
            filtered_df = filtered_df.assign(
                **{self._target: filtered_df[self._target].astype("int")}
            )

        self.estimator.fit(
            filtered_df[self._estimator_features], filtered_df[self._target]
        )

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")

        if hasattr(self.estimator, "predict_proba"):
            try:
                df[self._target] = df[self._target].astype("int")
            except Exception:
                pass

        df = self.transform_pre_transformers(df=df)
        if self.multiclassifier:
            df[self._pred_column] = self.estimator.predict_proba(
                df[self._estimator_features]
            ).tolist()
            df["classes"] = [list(self.estimator.classes_) for _ in range(len(df))]
            if len(set(df[self.pred_column].iloc[0])) == 2:
                raise ValueError(
                    "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly"
                )

        elif not hasattr(self._deepest_estimator, "predict_proba"):
            df[self._pred_column] = self.estimator.predict(df[self._estimator_features])
        else:
            df[self._pred_column] = self.estimator.predict_proba(
                df[self._estimator_features]
            )[:, 1]
        return df


class GranularityPredictor(BasePredictor):

    def __init__(
        self,
        granularity_column_name: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.granularity_column_name = granularity_column_name
        self.multiclassifier = multiclassifier
        self.column_names = column_names
        self._granularities = []
        self._granularity_estimators = {}

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100)"
            )

        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )
        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()

        filtered_df = apply_filters(df=df, filters=self.filters)
        if hasattr(self.estimator, "predict_proba"):
            try:
                filtered_df[self._target] = filtered_df[self._target].astype("int")
            except Exception:
                pass

        filtered_df = self.fit_transform_pre_transformers(df=filtered_df)

        if hasattr(self._deepest_estimator, "predict_proba"):
            filtered_df = filtered_df.assign(
                **{self._target: filtered_df[self._target].astype("int")}
            )

        if (
            not self.multiclassifier
            and len(filtered_df[self._target].unique()) > 2
            and hasattr(self._deepest_estimator, "predict_proba")
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)
            if len(filtered_df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets"
                )

        self._granularities = filtered_df[self.granularity_column_name].unique()

        for granularity in self._granularities:
            self._granularity_estimators[granularity] = clone(self.estimator)
            rows = filtered_df[filtered_df[self.granularity_column_name] == granularity]
            self._granularity_estimators[granularity].fit(
                rows[self._estimator_features], rows[self._target]
            )

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype("int")})
            except Exception:
                pass

        df = self.transform_pre_transformers(df=df)
        dfs = []
        for granularity, estimator in self._granularity_estimators.items():
            rows = df[df[self.granularity_column_name] == granularity]
            if self.multiclassifier:
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict_proba(
                            rows[self._estimator_features]
                        ).tolist()
                    }
                )
                rows = rows.assign(
                    **{"classes": [list(estimator.classes_) for _ in range(len(rows))]}
                )
                if len(set(rows[self.pred_column].iloc[0])) == 2:
                    raise ValueError(
                        "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly"
                    )

            elif not hasattr(self._deepest_estimator, "predict_proba"):
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict(
                            rows[self._estimator_features]
                        )
                    }
                )
            else:
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict_proba(
                            rows[self._estimator_features]
                        )[:, 1]
                    }
                )
            dfs.append(rows)

        df = pd.concat(dfs)
        return df


class SeriesWinLosePredictor(BasePredictor):
    def __init__(
        self,
        format_column_name: str,
        game_win_prob_column_name: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.format_column_name = format_column_name
        self.game_win_prob_column_name = game_win_prob_column_name
        self.multiclassifier = multiclassifier
        self.column_names = column_names
        win_filters = filters.copy()
        win_filters.append(
            Filter(
                column_name=self.game_win_prob_column_name,
                value=1,
                operator=Operator.EQUALS,
            )
        )
        lose_filters = filters.copy()
        lose_filters.append(
            Filter(
                column_name=self.game_win_prob_column_name,
                value=0,
                operator=Operator.EQUALS,
            )
        )

        self._win_predictor = Predictor(
            target=self._target,
            estimator=estimator,
            estimator_features=estimator_features,
            filters=win_filters,
            multiclassifier=multiclassifier,
            pred_column=pred_column,
            column_names=column_names,
            pre_transformers=pre_transformers,
        )
        self._lose_predictor = Predictor(
            target=self._target,
            estimator=estimator,
            estimator_features=estimator_features,
            filters=lose_filters,
            multiclassifier=multiclassifier,
            pred_column=pred_column,
            column_names=column_names,
            pre_transformers=pre_transformers,
        )

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100)"
            )

        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        self._win_predictor.train(df=df, estimator_features=estimator_features)
        self._lose_predictor.train(df=df, estimator_features=estimator_features)

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        win_df = df[df[self.game_win_prob_column_name] == 1]
        lose_df = df[df[self.game_win_prob_column_name] == 0]

        win_df = self._win_predictor.add_prediction(win_df)
        lose_df = self._lose_predictor.add_prediction(lose_df)

        return pd.concat([win_df, lose_df]).sort_values(
            by=[
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )


class PointToClassificationPredictor(BasePredictor):

    def __init__(
        self,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        point_estimate_column: Optional[str] = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.multiclassifier = multiclassifier
        self.point_estimate_column = point_estimate_column
        self.column_names = column_names
        self._target_probs = {}
        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMRegressor(
                max_depth=2, n_estimators=100, learning_rate=0.05, verbose=-100
            ),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        if self.point_estimate_column is not None:
            predictions = df[self.point_estimate_column]
        else:
            self.estimator.fit(df[estimator_features], df[self._target])
            predictions = self.estimator.predict(df[estimator_features])

        quantiles = predictions.quantile([q / 50 for q in range(1, 50)])

        for idx, quantile in enumerate(quantiles[:-1]):
            rows = df[(predictions >= quantile) & (predictions < quantiles[idx + 1])]
            value_counts = rows[self._target].value_counts()
            for target, count in value_counts:
                self._target_probs[target] = count / len(df)



from dataclasses import dataclass
from enum import Enum
from typing import Optional

import numpy as np
import pandas as pd
from player_performance_ratings.predictor import BasePredictor

from player_performance_ratings import ColumnNames

from player_performance_ratings.transformers.base_transformer import (
    BaseTransformer,
    BaseLagGenerator,
)


class NetOverPredictedPostTransformer(BaseTransformer):

    def __init__(
        self,
        predictor: BasePredictor,
        features: list[str] = None,
        lag_generators: Optional[list[BaseLagGenerator]] = None,
        prefix: str = "net_over_predicted_",
        are_estimator_features: bool = False,
    ):
        super().__init__(
            features=features,
            are_estimator_features=are_estimator_features,
            features_out=[],
        )
        self.prefix = prefix
        self.predictor = predictor
        self._features_out = []
        self.lag_generators = lag_generators or []
        self.column_names = None
        new_feature_name = self.prefix + self.predictor.pred_column
        self._features_out.append(new_feature_name)
        self._estimator_features_out = []
        for lag_generator in self.lag_generators:
            if not lag_generator.features:
                lag_generator.features = [new_feature_name]
                for iteration in lag_generator.iterations:
                    lag_generator._features_out = [
                        f"{lag_generator.prefix}{iteration}_{new_feature_name}"
                    ]
                    self.features_out.extend(lag_generator._features_out.copy())
                    self._estimator_features_out.extend(
                        lag_generator._features_out.copy()
                    )

        if self._are_estimator_features:
            self._estimator_features_out.append(self.predictor.pred_column)
            self.features_out.append(self.predictor.pred_column)
        if self.prefix is "":
            raise ValueError("Prefix must not be empty")

    def fit_transform(
        self, df: pd.DataFrame, column_names: Optional[ColumnNames] = None
    ) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        self.column_names = column_names
        self.predictor.train(df, estimator_features=self.features)
        df = self.predictor.add_prediction(df)
        new_feature_name = self.prefix + self.predictor.pred_column
        if self.predictor.target not in df.columns:
            df = df.assign(**{new_feature_name: np.nan})
        else:
            df = df.assign(
                **{
                    new_feature_name: df[self.predictor.target]
                    - df[self.predictor.pred_column]
                }
            )
        for lag_generator in self.lag_generators:
            df = lag_generator.generate_historical(df, column_names=self.column_names)
        return df[list(set(ori_cols + self.features_out))]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        df = self.predictor.add_prediction(df)
        new_feature_name = self.prefix + self.predictor.pred_column
        if self.predictor.target not in df.columns:
            df = df.assign(**{new_feature_name: np.nan})
        else:
            df = df.assign(
                **{
                    new_feature_name: df[self.predictor.target]
                    - df[self.predictor.pred_column]
                }
            )

        for lag_generator in self.lag_generators:
            df = lag_generator.generate_future(df)

        return df[list(set(ori_cols + self.features_out))]

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def reset(self) -> "BaseTransformer":
        for lag_generator in self.lag_generators:
            lag_generator.reset()
        return self


class Operation(Enum):
    SUBTRACT = "subtract"


@dataclass
class ModifyOperation:
    feature1: str
    operation: Operation
    feature2: str
    new_column_name: Optional[str] = None

    def __post_init__(self):
        if self.operation == Operation.SUBTRACT and not self.new_column_name:
            self.new_column_name = f"{self.feature1}_minus_{self.feature2}"


class ModifierTransformer(BaseTransformer):

    def __init__(
        self,
        modify_operations: list[ModifyOperation],
        features: list[str] = None,
        are_estimator_features: bool = True,
    ):
        self.modify_operations = modify_operations
        _features_out = [
            operation.new_column_name for operation in self.modify_operations
        ]
        super().__init__(
            features=features,
            are_estimator_features=are_estimator_features,
            features_out=_features_out,
        )

    def fit_transform(
        self, df: pd.DataFrame, column_names: Optional[ColumnNames]
    ) -> pd.DataFrame:
        self.column_names = column_names
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for operation in self.modify_operations:
            if operation.operation == Operation.SUBTRACT:
                if (
                    operation.feature1 not in df.columns
                    or operation.feature2 not in df.columns
                ):
                    df = df.assign(**{operation.new_column_name: np.nan})

                else:
                    df = df.assign(
                        **{
                            operation.new_column_name: df[operation.feature1]
                            - df[operation.feature2]
                        }
                    )

        return df


class PredictorTransformer(BaseTransformer):

    def __init__(self, predictor: BasePredictor, features: list[str] = None):
        self.predictor = predictor
        super().__init__(
            features=features, features_out=[f"{self.predictor.pred_column}"]
        )

    def fit_transform(
        self, df: pd.DataFrame, column_names: Optional[None] = None
    ) -> pd.DataFrame:
        self.predictor.train(df=df, estimator_features=self.features)
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)
        return df


class RatioTeamPredictorTransformer(BaseTransformer):
    def __init__(
        self,
        features: list[str],
        predictor: BasePredictor,
        team_total_prediction_column: Optional[str] = None,
        lag_generators: Optional[list[BaseLagGenerator]] = None,
        prefix: str = "_ratio_team",
    ):
        self.predictor = predictor
        self.team_total_prediction_column = team_total_prediction_column
        self.prefix = prefix
        self.predictor._pred_column = f"__prediction__{self.predictor.target}"
        self.lag_generators = lag_generators or []
        super().__init__(
            features=features,
            features_out=[self.predictor.target + prefix, self.predictor._pred_column],
        )

        if self.team_total_prediction_column:
            self._features_out.append(
                self.predictor.target + prefix + "_team_total_multiplied"
            )
        for lag_generator in self.lag_generators:
            self._features_out.extend(lag_generator.features_out)

        if self._are_estimator_features:
            self._estimator_features_out = self._features_out.copy()

    def fit_transform(
        self, df: pd.DataFrame, column_names: Optional[ColumnNames] = None
    ) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        self.column_names = column_names
        self.predictor.train(df=df, estimator_features=self.features)
        transformed_df = self.transform(df)
        for lag_generator in self.lag_generators:
            transformed_df = lag_generator.generate_historical(
                transformed_df, column_names=self.column_names
            )

        return transformed_df[list(set(ori_cols + self.features_out))]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)

        df[self.predictor.pred_column + "_sum"] = df.groupby(
            [self.column_names.match_id, self.column_names.team_id]
        )[self.predictor.pred_column].transform("sum")
        df[self._features_out[0]] = (
            df[self.predictor.pred_column] / df[self.predictor.pred_column + "_sum"]
        )
        if self.team_total_prediction_column:
            df = df.assign(
                **{
                    self.predictor.target
                    + self.prefix
                    + "_team_total_multiplied": df[self._features_out[0]]
                    * df[self.team_total_prediction_column]
                }
            )

        for lag_transformer in self.lag_generators:
            df = lag_transformer.generate_historical(df, column_names=self.column_names)

        return df.drop(columns=[self.predictor.pred_column + "_sum"])

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def reset(self) -> "BaseTransformer":
        for lag_generator in self.lag_generators:
            lag_generator.reset()
        return self
